# ==============================================================================
# Modules
# ==============================================================================
# Built-ins
import os, sys, requests, time

# Soothsayer Utils
from soothsayer.utils import add_objects_to_globals
import soothsayer_utils as syu
functions_from_soothsayer_utils = [
    "read_url",
]
add_objects_to_globals(syu, functions_from_soothsayer_utils, globals(), add_version=True, __all__=None)


# ======
# Future
# ======
# (1) Add ability for `path` arguments to be `pathlib.Path`
# (2) Add ability to process ~ in path
# =============
# Web
# =============
# # Reading html from website
# def read_url(url:str, params=None, **kwargs):
#     """
#     Future:
#     Make wrapper for dynamic html and phantom js
#     """
#     dynamic = False
#     if not dynamic:
#         r = requests.get(url, params=params, **kwargs)
#         return r.text
#     else:
#         print("Not finished. Need to make wrapper for dynamic HTML using PhantomJS", file=sys.stderr)



# ==============================================================================
# Imports
# ==============================================================================
# Built-ins
import os, sys, time, uuid

import pandas as pd
import numpy as np

# Soothsayer
from ..r_wrappers import *
from soothsayer.utils import check_packages

# from soothsayer.symmetry import *

# ==============================================================================
# R Imports
# ==============================================================================
if "rpy2" in sys.modules:
    from rpy2 import robjects as ro
    from rpy2 import rinterface as ri

    from rpy2.robjects.packages import importr
    try:
        from rpy2.rinterface import RRuntimeError
    except ImportError:
        from rpy2.rinterface_lib.embedded import RRuntimeError
    from rpy2.robjects import pandas2ri
    R = ro.r
    NULL = ri.NULL
    #rinterface.set_writeconsole_regular(None)

# R Packages
# phyloseq = R_package_retrieve("phyloseq")
# philr = R_package_retrieve("philr")
# ape = R_package_retrieve("ape")

# Phylogenetic Isometric log ratio
@check_packages(["phyloseq", "philr", "ape"], language="r", import_into_backend=False)
def normalize_philr(X:pd.DataFrame, tree, node_prefix="y", part_weights = "uniform", ilr_weights = "uniform", verbose=True, bifurcation_kws=dict(recursive=True)):
    """
    tree should be some type of ete3 Tree
    """
    # Imports
    phyloseq = R_package_retrieve("phyloseq")
    philr = R_package_retrieve("philr")
    ape = R_package_retrieve("ape")

    assert isinstance(X, pd.DataFrame), "type(df_counts) must be pd.DataFrame"
    assert not np.any(X == 0), "`X` cannot contain any zero values because of the log-transform.  Give it a pseudocount. (X+1)"

    # Prune tree
    tree_set =  set(tree.get_leaf_names())
    leaf_set_from_X = set(X.columns)
    assert leaf_set_from_X <= tree_set, "`X` columns should be a subset of `tree` leaves"
    tree = tree.copy(method="deepcopy")
    if leaf_set_from_X < tree_set:
        n_before_pruning = len(tree_set)
        tree.prune(leaf_set_from_X)
        if verbose:
            print(f"Pruned {n_before_pruning - len(tree.get_leaves())} attributes to match X.columns", file=sys.stderr)

    # Check bifurcation
    n_internal_nodes = len([*filter(lambda node:node.is_leaf() == False, tree.traverse())])
    n_leaves = len([*filter(lambda node:node.is_leaf(), tree.traverse())])
    if n_internal_nodes < (n_leaves - 1):
        tree.resolve_polytomy(**bifurcation_kws)

    # Convert to R datatype
    r_df_counts = pandas_to_rpy2(X)
    # I/O into PhyloSeq object
    path_newick_temporary = uuid.uuid4().hex
    tree.write(format=1, outfile=path_newick_temporary)
    r_tree_phyloseq = phyloseq.read_tree(path_newick_temporary)
    os.remove(path_newick_temporary)
    # Relabel
    r_tree_phyloseq = ape.makeNodeLabel(r_tree_phyloseq, method="number", prefix=node_prefix)
    # Transform
    r_df_transform = philr.philr(df=R["as.matrix"](r_df_counts), tree=r_tree_phyloseq, part_weights=part_weights, ilr_weights=ilr_weights)
    return pd.DataFrame(rpy2_to_pandas(r_df_transform), index=X.index, columns=list(map(lambda x:f"{node_prefix}{x}", range(1,X.shape[1]))))



# ==============================================================================
# Modules
# ==============================================================================
# Soothsayer
import soothsayer_utils as syu
from soothsayer.utils import add_objects_to_globals

functions_from_soothsayer_utils = [
'get_iris_data',
]
add_objects_to_globals(syu, functions_from_soothsayer_utils, globals(), add_version=True, __all__=None)



# ==============================================================================
# Modules
# ==============================================================================
# Built-ins
import os, sys, time

# ===
# I/O
# ===
from .bioinformatics import *
from .datasets import *
from .objects import *
from .tables import *
from .text import *
from .web import *

_bioinformatics = [
"read_clustalo_distmat", "read_vizbin_coords", "read_checkm_qa", "read_fasta", 
"write_fasta", "read_blast", "read_gtf_gff_base", "read_gff3", "read_gtf", "read_ncbi_xml", "read_ebi_sample_metadata",
"read_star_log", "read_hmmer",
]
_datasets = ["get_iris_data"]
_objects = ["write_object", "read_object", "read_script_as_module"]
_tables = ["read_dataframe", "write_dataframe", "read_multiple_synopsis"]
_text = ["read_textfile", "get_file_object", "read_from_clipboard"]
_web = ["read_url"]


__all__ = _bioinformatics + _datasets + _objects + _tables + _text + _web
__all__ = sorted(__all__)



# ==============================================================================
# Modules
# ==============================================================================
# Built-ins
import os, sys, time, copy, pathlib, importlib
from collections import *

# Compression & Serialization
import pickle, gzip, bz2, zipfile

# Soothsayer
from soothsayer.utils import format_path, infer_compression, add_objects_to_globals

# Soothsayer Utils
import soothsayer_utils as syu
functions_from_soothsayer_utils = [
    "read_object",
    "write_object",
    "read_script_as_module",
]
add_objects_to_globals(syu, functions_from_soothsayer_utils, globals(), add_version=True, __all__=None)


# ======
# Future
# ======
# (1) Add ability for `path` arguments to be `pathlib.Path`
# (2) Add ability to process ~ in path

# =============
# Serialization
# =============
# # Writing serial object
# def write_object(obj, path:str, compression="infer", serialization_module=pickle, protocol=pickle.HIGHEST_PROTOCOL, *args):
#     """
#     Extensions:
#     pickle ==> .pkl
#     dill ==> .dill
#     gzipped-pickle ==> .pgz
#     bzip2-pickle ==> .pbz2
#     """
#     assert obj is not None, "Warning: `obj` is NoneType"
#     path = format_path(path, str)

#     # Use infer_compression here
#     if compression == "infer":
#         _ , ext = os.path.splitext(path)
#         if (ext == ".pkl") or (ext == ".dill"):
#             compression = None
#         if ext in {".pgz", ".gz"}:
#             compression = "gzip"
#         if ext in {".pbz2", ".bz2"}:
#             compression = "bz2"
#     if compression is not None:
#         if compression == "bz2":
#             f = bz2.BZ2File(path, "wb")
#         if compression == "gzip":
#             f = gzip.GzipFile(path, "wb")
#     else:
#         f = open(path, "wb")
#     serialization_module.dump(obj, f, protocol=protocol, *args)
#     f.close()

# # Reading serial object
# def read_object(path:str, compression="infer", serialization_module=pickle):
#     path = format_path(path, str)

#     if compression == "infer":
#         _ , ext = os.path.splitext(path)
#         if (ext == ".pkl") or (ext == ".dill"):
#             compression = None
#         if ext in {".pgz", ".gz"}:
#             compression = "gzip"
#         if ext in {".pbz2", ".bz2"}:
#             compression = "bz2"
#     if compression is not None:
#         if compression == "gzip":
#             f = gzip.open(path, "rb")
#         if compression == "bz2":
#             f = bz2.open(path, "rb")
#     else:
#         f = open(path, "rb")
#     obj = serialization_module.load(f)
#     f.close()
#     return obj

# # Importing a functions from a module
# def read_script_as_module(name_module, path):
#     # https://stackoverflow.com/questions/67631/how-to-import-a-module-given-the-full-path
#     path = format_path(path, str)
#     spec = importlib.util.spec_from_file_location(name_module, path)
#     module = importlib.util.module_from_spec(spec)
#     spec.loader.exec_module(module)
#     return module




        
# ==============================================================================
# Modules
# ==============================================================================
# Built-ins
import os, sys, time, warnings

# PyData
import numpy as np
import pandas as pd
from scipy.spatial.distance import euclidean 

# Scikits
from sklearn.preprocessing import StandardScaler
from sklearn.decomposition import PCA
from sklearn.manifold import TSNE
from skbio.stats.ordination import pcoa, pcoa_biplot
from skbio import DistanceMatrix
from adjustText import adjust_text

from ..io import write_object
from ..utils import is_query_class, to_precision, dict_filter, is_symmetrical, is_nonstring_iterable, assert_acceptable_arguments
from ..symmetry import Symmetric
from ..visuals import plot_scatter

from matplotlib.patches import ConnectionPatch
import matplotlib.pyplot as plt

__all__ = ["MatrixDecomposition", "PrincipalComponentAnalysis","PrincipalCoordinatesAnalysis", "Manifold", "eigenprofiles_from_data"]
__all__ = sorted(__all__)


# General ordination methods
class CoreOrdinationMethods(object):
    # Get principal components
    def __getitem__(self, key):
        if type(key) == int:
            key = f"{prefix}{key}"
        return self.projection_[key]

        
    def _plot_biplot(
            self, 
            ax, 
            show_features,  
            precision, 
            use_percent, 
            method_top_features,
            arrowstyle,
            arrow_scale,
            color_arrow, 
            color_feature, 
            show_origin, 
            adjust_label_positions,
            autoscale_xpad, 
            autoscale_ypad, 
            arrow_kws, 
            arrow_properties,
            text_kws,
            **plot_scatter_kws,
            ):
            """
            Adapted from `erdogant/pca` on GitHub:
            https://github.com/erdogant/pca/blob/47d3dd101f6c757d783159f4f084f0dd38bf0412/pca/pca.py#L431
            to use skbio OrdinationResults object

            Create the Biplot based on model.
            
            Returns
            -------
            ax
            """
            warnings.warn("Using `show_features` to create a `biplot` is experimental.")
            
            # Pre-processing
            if show_features is True:
                show_features = -1
            if isinstance(show_features, int):
                assert show_features <= self.m, "`show_features` is larger than the number of features available.  Please use -1 to use all features."
                if show_features == -1:
                    # Set -1 to total number of features
                    show_features = self.m
                    
                # Get top features
                assert_acceptable_arguments(method_top_features, {"weighted", "partitioned", "importance"})
                number_of_requested_features = show_features
                if method_top_features == "weighted":
                    weighted_features = (self.contributions_.iloc[:,:2]*self.explained_variance_ratio_.iloc[:2]).sum(axis=1).sort_values(ascending=False)
                    show_features = weighted_features.iloc[:number_of_requested_features].index
                if method_top_features == "partitioned":
                    number_of_requested_features_is_even = (number_of_requested_features % 2) == 0
                    k = number_of_requested_features//2
                    # Even
                    if number_of_requested_features_is_even: 
                        pc1_features = self.contributions_.iloc[:,0].sort_values(ascending=False).iloc[:k].index.tolist()
                    # Odd
                    else:
                        pc1_features = self.contributions_.iloc[:,0].sort_values(ascending=False).iloc[:k+1].index.tolist()
                    pc2_features = self.contributions_.iloc[:,1].sort_values(ascending=False).iloc[:k].index.tolist()
                    show_features = sorted(set(pc1_features + pc2_features))
                if method_top_features == "importance":
                    # https://github.com/qiime2/q2-emperor/blob/756eddba1fc590f33d4bf7a3d0268517e68e9ae5/q2_emperor/_plot.py#L78
                    # select the top N most important features based on the vector's magnitude
                    show_features = self.importance_.iloc[:number_of_requested_features].index
                print("Selected {} features:".format(number_of_requested_features), list(show_features), sep="\n", file=sys.stderr)
                
            if isinstance(show_features, str):
                show_features = [show_features]
            if is_nonstring_iterable(show_features):
                assert set(show_features) <= set(self.loadings_.index), "`show_features` must be either an integer <= m (-1 to include all) or a feature subset of `data.columns`"
                
            # Get loading subset
            loadings = self.loadings_.loc[show_features,:].iloc[:,:2]

            # Pre-processing
            # Use the PCs only for scaling purposes
            x = self.projection_.iloc[:,0].values
            y = self.projection_.iloc[:,1].values

            mean_x = np.mean(x)
            max_x = np.abs(x).max()
            mean_y = np.mean(y)
            max_y = np.abs(y).max()

            # Keywords
            _arrow_kws = dict(color=color_arrow, linewidth=1, alpha=0.618) #, head_width=0.003, head_length = 0.005, alpha=0.382,  shape="left", overhang = 2.0, length_includes_head=False) #  head_width=0.1,
            _arrow_kws.update(arrow_kws)
            _arrowhead_kws = dict(arrowstyle=arrowstyle, mutation_scale=20, facecolor=color_arrow, alpha=0.618)
            
            _text_kws = dict(color=color_feature, ha='center', va='center')
            _text_kws.update(text_kws)

            # Plot arrows and text
            contributions = self.contributions_.loc[show_features,:].iloc[:,:2]
            text_objects = list()
            for i in range(0, loadings.shape[0]):
                feature = loadings.index[i]
                # Set PC1 vs PC2 direction. Note that these are not neccarily the best loading.
                xarrow = contributions.loc[feature].iloc[0] * np.sign(loadings.loc[feature].iloc[0]) * max_x * arrow_scale
                yarrow = contributions.loc[feature].iloc[1] * np.sign(loadings.loc[feature].iloc[1]) * max_y * arrow_scale

                # Plot arrow
                xyA = (mean_x, mean_y)
                xyB = (xarrow, yarrow)
                con = ConnectionPatch(xyA, xyB, coordsA="data", coordsB="data", **_arrowhead_kws)
                ax.plot([xyA[0], xyB[0]], [xyA[1], xyB[1]], **_arrow_kws)
                ax.add_artist(con)

                # Plot feature label
                text_objects.append(ax.text(xarrow*1.11, yarrow*1.11, feature,  **_text_kws))
                
            # Plot origin
            if show_origin:
                ax.scatter([mean_x], [mean_y], c=color_arrow, s=15)
                
            # Adjust label positions
            if adjust_label_positions:
                adjust_text(text_objects,  ax=ax)
                
            # Autoscale padding
            if autoscale_xpad is None:
                autoscale_xpad = 0
            if autoscale_ypad is None:
                autoscale_ypad = 0
                
            if any([autoscale_xpad, autoscale_ypad]):
                if bool(autoscale_xpad):
                    assert autoscale_xpad < 1
                    ax.set_xlim(*np.asarray(ax.get_xlim())*(1 + autoscale_xpad))
                if bool(autoscale_ypad):
                    assert autoscale_ypad < 1
                    ax.set_ylim(*np.asarray(ax.get_ylim())*(1 + autoscale_ypad))
                    
            return ax
        
    # Plotting
    def plot(
        self,    
        ax=None, 
        show_features=None,  
        precision=4, 
        use_percent=False, 
        method_top_features="partitioned",
        arrowstyle="-|>",
        arrow_scale=1,
        color_arrow="darkslategray", 
        color_feature="darkslategray",
        show_xaxis=True,
        show_yaxis=True,
        show_origin=True, 
        adjust_label_positions=False,
        autoscale_xpad=None, 
        autoscale_ypad=None, 
        arrow_kws=dict(), 
        arrow_properties=dict(),
        text_kws=dict(),
        axis_kws=dict(),
        **plot_scatter_kws,
        ):
        """
        See soothsayer.visuals.plot_scatter for details on parameters.
        """
        # Update scatter_kws
        plot_scatter_kws["ax"] = ax
        
        xlabel = f"{self.prefix}1"
        ylabel = f"{self.prefix}2"
        if hasattr(self, "explained_variance_ratio_"):
            if not use_percent:
                ev1 = to_precision(self.explained_variance_ratio_[f"{self.prefix}1"], precision=precision)
                ev2 = to_precision(self.explained_variance_ratio_[f"{self.prefix}2"], precision=precision)
                xlabel = f"{xlabel}({ev1})"
                ylabel = f"{ylabel}({ev2})"
            else:
                ev1 = to_precision(self.explained_variance_ratio_[f"{self.prefix}1"]*100, precision=precision)
                ev2 = to_precision(self.explained_variance_ratio_[f"{self.prefix}2"]*100, precision=precision)
                xlabel = f"{xlabel}({ev1}%)"
                ylabel = f"{ylabel}({ev2}%)" 
                
        # Plot scatter
        fig, ax, im = plot_scatter(data=self.projection_, x=f"{self.prefix}1", y=f"{self.prefix}2", xlabel=xlabel, ylabel=ylabel, **plot_scatter_kws)
        
        # Plot features (biplot)
        if not any([(show_features is False), (show_features is None)]):
            if hasattr(self, "loadings_"):
                assert self.loadings_ is not None, "Cannot visualize biplot for ordination objects without loadings.  If using `PrincipalCoordinatesAnalysis`, please rerun with `data` argument."
                # Plot biplot
                self._plot_biplot(
                    ax=ax,
                    show_features=show_features,
                    precision=precision,
                    use_percent=use_percent,
                    method_top_features=method_top_features,
                    arrowstyle=arrowstyle,
                    arrow_scale=arrow_scale,
                    color_arrow=color_arrow, 
                    color_feature=color_feature, 
                    show_origin=show_origin, 
                    adjust_label_positions=adjust_label_positions,
                    autoscale_xpad=autoscale_xpad, 
                    autoscale_ypad=autoscale_ypad, 
                    arrow_kws=arrow_kws, 
                    arrow_properties=arrow_properties,
                    text_kws=text_kws,
                )
            else:
                warnings.warn("Cannot visualize biplot for ordination objects without loadings")
                
        # Show axis
        _axis_kws = {"linestyle":"-", "color":"black", "linewidth":0.618, "alpha":0.1618}
        _axis_kws.update(axis_kws)
        if show_xaxis:
            ax.axhline(0, **_axis_kws)
        if show_yaxis:
            ax.axvline(0, **_axis_kws)
        return fig, ax, im
    

    # Plot Scree
    def plot_scree(
        self, 
        title=None, 
        target_explained_variance=None, 
        alpha_barchart=0.618, 
        alpha_marker=0.618, 
        alpha_target=0.618, 
        color_barchart="teal", 
        color_marker="black", 
        color_target="maroon",  
        show_xgrid=True, 
        show_ygrid=False, 
        fig_kws=dict(), 
        label_kws=dict(), 
        target_kws=dict(), 
        title_kws=dict(), 
        legend_kws=dict(),
        show_n_components=None, 
        ax=None, 
        bar_kws=dict(), 
        tick_kws=dict(), 
        line_kws=dict(), 
        style="seaborn-white",
        precision=4, 
        ):
        # Keywords
        _fig_kws = {"figsize":(8,3)}
        _fig_kws.update(fig_kws)
        _label_kws = {"fontsize":15}
        _label_kws.update(label_kws)
        _bar_kws = {"edgecolor":"black", "linewidth":1.618, "color":color_barchart, "alpha":alpha_barchart}
        _bar_kws.update(bar_kws)
        _line_kws = {"marker":"o",  "markeredgewidth":1, "markeredgecolor":"black", "color":color_marker, "alpha":alpha_marker}
        _line_kws.update(line_kws)
        _target_kws={"alpha":alpha_target, "color":color_target, "linestyle":":"}
        _target_kws.update(target_kws)
        _tick_kws = {"fontsize":12}
        _tick_kws.update(tick_kws)
        _title_kws = {"fontsize":15, "fontweight":"bold"}
        _title_kws.update(title_kws)
        _legend_kws = {"fontsize":12}
        _legend_kws.update(legend_kws)

        with plt.style.context(style):
            if ax is None:
                fig, ax = plt.subplots(**_fig_kws)
            else:
                fig = plt.gcf()
            # Plotting
            evr = self.explained_variance_ratio_
            evr[:show_n_components].plot(kind="bar",  ax=ax, label='_nolegend_', **_bar_kws)
            evr_cumsum = self.explained_variance_ratio_.cumsum()
            evr_cumsum[:show_n_components].plot(kind="line",  label='_nolegend_', ax=ax, **_line_kws)

            if target_explained_variance is not None:
                assert isinstance(target_explained_variance, float)
                idx_at_target = (evr_cumsum >= target_explained_variance).idxmax()
                if target_explained_variance:
                    x_pos = evr_cumsum.index.get_loc(idx_at_target)
                    y_pos = evr_cumsum[idx_at_target]
                    ax.plot([x_pos]*2, [0, y_pos], label="EV({}) ≥ {} [{}]".format(idx_at_target, to_precision(target_explained_variance, precision=precision), to_precision(y_pos, precision=precision)), **_target_kws)
                    ax.plot([min(ax.get_xlim()), x_pos], [y_pos, y_pos], **_target_kws)
                    ax.legend(**_legend_kws)


            # References
            ax.axhline(1.0, color="black", linestyle="-", linewidth=1)
            ax.set_ylabel("Explained Variance", **_label_kws)
            ax.set_xticklabels(ax.get_xticklabels(), **_tick_kws)

            if title is not None:
                ax.set_title(title, **_title_kws)
            ax.xaxis.grid(show_xgrid)
            ax.yaxis.grid(show_ygrid)

            ax.set_ylim((0,1.06180339887))

            return fig, ax
        
    def plot_feature_contribution(
        self,
        cmap=plt.cm.magma_r,
        vmin=0, 
        vmax=1,
        annot=True,
        linecolor="white",
        linewidth=1,
        title=None,
        show_eigenvectors=-1,
        show_features=-1,
        cbar_label="Explained variance ratio",
        ax=None,
        style="seaborn-white",
        fig_kws=dict(),
        heatmap_kws=dict(),
        title_kws=dict(fontsize=15, fontweight="bold")
        ):
        assert hasattr(self,"contributions_")
        assert self.contributions_ is not None
        if show_eigenvectors == -1:
            show_eigenvectors = self.contributions_.columns
        if show_features == -1:
            show_features = self.contributions_.index
        _heatmap_kws = dict(cmap=cmap, vmin=vmin, vmax=vmax, annot=annot, linecolor=linecolor, linewidth=linewidth, cbar_kws=dict(label=cbar_label))
        _heatmap_kws.update(heatmap_kws)
        with plt.style.context(style):
            if ax is None:
                fig, ax = plt.subplots(**fig_kws)
            else:
                fig = plt.gcf()
            sns.heatmap(self.contributions_.loc[show_features, show_eigenvectors], ax=ax, **_heatmap_kws)
            if title:
                ax.set_title(title, **title_kws)
        return ax
        

    # Save to file
    def to_file(self, path:str, compression="infer"):
        write_object(self, path, compression=compression)
        

# Sklearn Decomposition wrapper class
class MatrixDecomposition(object):
    """
    Wrapper for sklearn.decomposition objects
    https://scikit-learn.org/stable/modules/classes.html#module-sklearn.decomposition
    """
    def __init__(self, data:pd.DataFrame, center=True, scale=False, n_components=-1, base_model=PCA, model_kws=dict(), prefix="PC.", name=None):
        self.data = data
#         self.data_transformed_ = data.copy()
        self.n, self.m = data.shape
        self.center = center
        self.scale = scale
        self.name = name
        self.prefix = prefix
        self.n_components = self._check_n_components(*data.shape, n_components)

        # Preprocess
        self.data_transformed_ = pd.DataFrame(
            data=StandardScaler(copy=True, with_mean=center, with_std=scale).fit_transform(data),
            index=data.index,
            columns=data.columns,
        )

        # Model
        self.model_ = self._build_model(base_model=base_model, data=self.data_transformed_, model_kws=model_kws)

        # Explained Variance
        if hasattr(self.model_, "explained_variance_ratio_"):
            self.explained_variance_ratio_ = pd.Series(
                self.model_.explained_variance_ratio_, 
                name="Explained variance ratio",
            )
            self.explained_variance_ratio_.index = self.explained_variance_ratio_.index.map(lambda i: f"{prefix}{i+1}")

        # Loadings
        if hasattr(self.model_, "components_"):
            self.loadings_ = pd.DataFrame(
                self.model_.components_.T, 
                index=self.data_transformed_.columns,
            )
            self.loadings_.columns = self.loadings_.columns.map(lambda i: f"{prefix}{i+1}")
            
        # Contributions
        if hasattr(self, "loadings_"):
            self.contributions_ = self.loadings_.abs()
            self.contributions_ = self.contributions_/self.contributions_.sum(axis=0)

        # Importance
        # https://github.com/qiime2/q2-emperor/blob/master/q2_emperor/_plot.py#L78
        if hasattr(self, "loadings_"):
            if self.loadings_ is not None:
                origin = np.zeros(self.loadings_.shape[1])
                self.importance_ = self.loadings_.apply(euclidean, axis=1, args=(origin,)).sort_values(ascending=False)
            else:
                self.importance_ = None
                
        # Project data onto eigenvectors
        self.projection_ =  pd.DataFrame(
            self.model_.transform(self.data_transformed_),
            index = self.data_transformed_.index,
        )
        self.projection_.columns = self.projection_.columns.map(lambda i: f"{prefix}{i+1}")

    # Check number of possible components
    def _check_n_components(self, n, m, n_components):
        possible_components = min(n, m)
        if n_components == -1:
            n_components = possible_components
        if (n_components > possible_components):
            n_components = possible_components
        return n_components

    # Construct model
    def _build_model(self, base_model, data, model_kws):
        model = base_model(n_components=self.n_components, **model_kws)
        model.fit(data)
        return model

# Principal Component Analysis
class PrincipalComponentAnalysis(CoreOrdinationMethods):
    """
    See soothsayer.ordination.MatrixDecomposition for details on parameters.
    https://scikit-learn.org/stable/modules/generated/sklearn.decomposition.PCA.html#sklearn.decomposition.PCA
    """
    def __init__(self, data:pd.DataFrame, center=True, scale=False, n_components=-1, base_model=PCA,  model_kws=dict(svd_solver="full", whiten=False), prefix="PC.", name=None):
        self.core_ = MatrixDecomposition(
            data=data,
            center=center,
            scale=scale,
            n_components=n_components,
            base_model=base_model,
            model_kws=model_kws,
            prefix=prefix,
            name=name,
        )
        for attr,obj in self.core_.__dict__.items():
            setattr(self, attr, obj)


# Principal Coordinant Analysis
class PrincipalCoordinatesAnalysis(CoreOrdinationMethods):
    """
    Wrapper for skbio.stats.ordination.pcoa
    http://scikit-bio.org/docs/latest/generated/generated/skbio.stats.ordination.pcoa.html#skbio.stats.ordination.pcoa
    """
    
    # Initialize
    def __init__(self, dism, data:pd.DataFrame=None,name=None, method='eigh', metric_type=None, node_type=None, prefix="PCoA."):
        if isinstance(dism, pd.DataFrame):
            dism = Symmetric(dism, name=name, edge_type=metric_type, association="dissimilarity")
        self.dism = dism
        if data is not None:
            assert set(dism.nodes) == set(data.index), "`dism` must have same sample set as `data.index`"
            data = data.loc[dism.nodes]
            self.n, self.m = data.shape
        self.data = data
        self.labels = dism.nodes
        self.node_type = node_type
        self.metric_type = metric_type
        self.name = name
        self.method = method
        self.prefix = prefix

        # Model
        self.model_ = self._build_model(dism=dism, data=data, base_model=pcoa)

        # Explained Variance
        if hasattr(self.model_, "proportion_explained"):
            self.explained_variance_ratio_ = pd.Series(
                self.model_.proportion_explained.values, 
                name="Explained variance ratio",
            )
            self.explained_variance_ratio_.index = self.explained_variance_ratio_.index.map(lambda i: f"{prefix}{i+1}")

        # Components
        if hasattr(self.model_, "eigvals"):
            self.eigvals_ = pd.Series(
                self.model_.eigvals.values, 
                name="Eigenvalues",
            )
            self.eigvals_.index = self.eigvals_.index.map(lambda i: f"{prefix}{i+1}")

        # Loadings
        if hasattr(self.model_, "features"):
            if self.model_.features is not None:
                self.loadings_ = self.model_.features.iloc[:,:self.m]
                self.loadings_.columns = self.loadings_.columns.map(lambda x: "{}{}".format(self.prefix, x[2:]))
            else:
                self.loadings_ = None
                
        # Contributions
        if hasattr(self, "loadings_"):
            if self.loadings_ is not None:
                self.contributions_ = self.loadings_.abs()
                self.contributions_ = self.contributions_/self.contributions_.sum(axis=0)
            else:
                self.contributions_ = None
                
        # Importance
        # https://github.com/qiime2/q2-emperor/blob/master/q2_emperor/_plot.py#L78
        if hasattr(self, "loadings_"):
            if self.loadings_ is not None:
                origin = np.zeros(self.loadings_.shape[1])
                self.importance_ = self.loadings_.apply(euclidean, axis=1, args=(origin,)).sort_values(ascending=False)
            else:
                self.importance_ = None
                
        # Project data onto eigenvectors
        self.projection_ =  pd.DataFrame(
            self.model_.samples.values,
            index = self.labels,
        )
        self.projection_.columns = self.projection_.columns.map(lambda i: f"{prefix}{i+1}")

    # Construct model
    def _build_model(self, dism, data, base_model):
        distance_matrix = DistanceMatrix(data=dism.values, ids=dism.nodes)
        ordination_results = base_model(distance_matrix)
        if data is not None:
            ordination_results = pcoa_biplot(ordination_results, data)
        return ordination_results

# Manifold Learning
class Manifold(CoreOrdinationMethods):
    """
    Wrapper for sklearn.manifold objects
    https://scikit-learn.org/stable/modules/classes.html#module-sklearn.manifold
    """
    def __init__(self, X:pd.DataFrame, base_model=TSNE, n_components=2, compute_pca="infer", n_components__pca=50, pca_kws=dict(center=True, scale=False), random_state=0, prefix="infer", name=None, **model_kws):
        # Init
        base_model_name = str(base_model)[:-2].split(".")[-1]
        self.name = name
        if prefix == "infer":
            if base_model_name == "TSNE":
                prefix = f"t-SNE."
            else:
                prefix = f"{base_model_name}."
        self.prefix = prefix

        # Is X precomputed?
        self.X = X.copy()
        self.X_input = self.X
        self.precomputed = self._check_precomputed(self.X, model_kws)
        if self.precomputed:
            compatible_for_precomputed = any(x == base_model_name for x in ["TSNE",  "MDS", "SpectralEmbedding", "spectral_embedding", "UMAP"])
            assert compatible_for_precomputed, f"{base_model_name} is not compatible for precomputed option"
            if base_model_name == "TSNE":
                model_kws["metric"] = "precomputed"
            if base_model_name == "MDS":
                model_kws["dissimilarity"] = "precomputed"
            if base_model_name == "SpectralEmbedding":
                model_kws["affinity"] = "precomputed"
            if base_model_name == "spectral_embedding":
                model_kws["adjacency"] = "precomputed"
        else:
            # Should PCA be computed?
            if compute_pca == "infer":
                if base_model_name == "TSNE":
                    compute_pca = True
            if compute_pca:
                # Check how many PCA components can be computed from the data
                self.n_components__pca = self._check_n_components(*self.X.shape, n_components=n_components__pca)

                self.X_input = PrincipalComponentAnalysis(data=self.X, base_model=PCA, n_components=self.n_components__pca, **pca_kws).projection_

        # Random state
        self.random_state = random_state
        compatible_for_rs = ["LocallyLinearEmbedding", "TSNE", "MDS", "SpectralEmbedding", "locally_linear_embedding", "smacof", "spectral_embedding", "UMAP"]
        if base_model_name in compatible_for_rs:
            model_kws["random_state"] = self.random_state

        # Check number of components
        self.n_components = self._check_n_components(*self.X_input.shape, n_components)


         # Model
        self.model = self._build_model(base_model=base_model, X_input=self.X_input, model_kws=model_kws)
        self.__dict__.update(dict_filter(self.model.__dict__, keys=lambda x:x.endswith("_")))

        # Get projections/embeddings
        if hasattr(self, "embedding_"):
            self.projection_ =  pd.DataFrame(
            self.embedding_,
            index = self.X_input.index,
        )
        self.projection_.columns = self.projection_.columns.map(lambda i: f"{prefix}{i+1}")


    # Check number of possible components
    def _check_n_components(self, n, m, n_components):

        possible_components = min(n, m)
        if n_components == -1:
            n_components = possible_components
        if (n_components > possible_components):
            warnings.warn(f"Data shape could not support {n_components} and is using the maximum {possible_components} instead.")
            n_components = possible_components
        return n_components

    # Check if data is precomputed
    def _check_precomputed(self, X, model_kws):
        # Preprocess
        precomputed = None
        if "metric" in model_kws:
            if model_kws["metric"] == "precomputed":
                precomputed = True
            else:
                precomputed = False
        if precomputed is None:
            if is_query_class(X, "Symmetric"):
                X = X.to_dense()
                precomputed = True
            else:
                precomputed = is_symmetrical(X, tol=1e-5)
        return precomputed

    # Construct model
    def _build_model(self, base_model, X_input, model_kws):
        model = base_model(n_components=self.n_components, **model_kws)
        model.fit(X_input)
        return model

    # Make compatible with scree plotting
    def plot_scree(self, **args):
        print("Cannot create scree plot for Manifolds", file=sys.stderr)
        return self

# =======================
# Wraps on wraps on wraps
# =======================
# Get eigenprofiles from data
def eigenprofiles_from_data(X:pd.DataFrame, y:pd.Series, class_type:str = None, mode="eigenprofiles", pca_kws=dict(),  axis=0, prefix="PC."):
    """
    Calculate eigenprofiles from data
    2018-July-13
    """
    accepted_modes = ["eigenprofiles", "eigenvalues", None]
    assert mode in accepted_modes, f"'{mode}' is not in accepted modes: {accepted_modes}"
    if axis == 1:
        X = X.T.copy()
    idx_attributes = X.columns
    profiles = list()
    d_eigenvalues = dict()

    if mode is None:
        tasks = ["eigenprofiles", "eigenvalues"]
    else:
        tasks = [mode]
    for group, _X in X.groupby(y, axis=0):
        _X_T = _X.T
        pca_model = PrincipalComponentAnalysis(_X_T, n_components=1, prefix=prefix, **pca_kws)
        if "eigenprofiles" in tasks:
            pc1 = pca_model.projection_[f"{prefix}1"]
            eigenprofile = pd.Series(pc1, index=idx_attributes, name=group)
            profiles.append(eigenprofile)
        if "eigenvalues" in tasks:
            eigenvalue = pca_model.explained_variance_ratio_[f"{prefix}1"]
            d_eigenvalues[group] = eigenvalue
    output = list()
    if "eigenprofiles" in tasks:
        df_eigenprofiles = pd.DataFrame(profiles)
        df_eigenprofiles.index.name = class_type
        output.append(df_eigenprofiles)
    if "eigenvalues" in tasks:
        eigenvalues = pd.Series(d_eigenvalues, name=class_type)
        output.append(eigenvalues)
    assert len(output) > 0, "Check `X`, `y`, and `mode` because no computation was done"
    if len(output) == 1:
        return output[0]
    else:
        return tuple(output)



import os,sys,operator, warnings
import numpy as np
import pandas as pd
import networkx as nx
from scipy.stats import entropy
import matplotlib.pyplot as plt
from matplotlib.colors import is_color_like
from matplotlib.patches import Ellipse
from mpl_toolkits.axes_grid1.axes_divider import make_axes_locatable
from matplotlib.collections import LineCollection, PatchCollection
import matplotlib.transforms as mtrans
import matplotlib_venn
import seaborn as sns
from adjustText import adjust_text
from collections import OrderedDict, defaultdict
from ..microbiome import prevalence
from ..utils import *

__all__ = ["OutlineCollection", "draw_networkx_labels_with_box", "plot_scatter", "plot_venn_diagram", "plot_waterfall", "plot_volcano", "plot_annotation", "plot_prevalence", "plot_compositional", "bezier_points"]
__all__ = sorted(__all__)

# Bezier points for arc plots
def bezier_points(p1, p2, control, granularity=20):
    """
    Credit: William Hedley Thompson
    https://teneto.readthedocs.io/en/latest/_modules/teneto/plot/slice_plot.html#slice_plot

    This has been adapted from the teneto package.

    # Following 3 Function that draw vertical curved lines from around points.
    # p1 nad p2 are start and end trupes (x,y coords) and pointN is the resolution of the points
    # negxLim tries to restrain how far back along the x axis the bend can go.
    """
    def pascal_row(n):
        # This returns the nth row of Pascal's Triangle
        result = [1]
        x, numerator = 1, n
        for denominator in range(1, n // 2 + 1):
            x *= numerator
            x /= denominator
            result.append(x)
            numerator -= 1
        if n & 1 == 0:
            # n is even
            result.extend(reversed(result[:-1]))
        else:
            result.extend(reversed(result))
        return np.asarray(result)
    # These two functions originated from the plot.ly's documentation for python API.
    # They create points along a curve.
    def make_bezier(points):
        # xys should be a sequence of 2-tuples (Bezier control points)
        n = len(points)
        combinations = pascal_row(n - 1)

        def bezier(ts):
            # This uses the generalized formula for bezier curves
            # http://en.wikipedia.org/wiki/B%C3%A9zier_curve#Generalization
            result = []
            for t in ts:
                tpowers = np.power(np.ones_like(n)*t, np.arange(n))
                upowers = np.power(np.ones_like(n)*(1-t), np.arange(n))[::-1]
                coefs = combinations*tpowers*upowers

                result.append(tuple(np.sum(coefs*ps) for ps in zip(*points))) # Need to optimize this part
            return np.asarray(result)
        return bezier

    ts = np.arange(granularity+1)/granularity
    d = p1[0] - (max(p1[1], p2[1]) - min(p1[1], p2[1])) / control
    points = np.asarray([p1, (d, p1[1]), (d, p2[1]), p2])
    bezier = make_bezier(points)
    points = bezier(ts)
    bvx = points[:,0]
    bvy = points[:,1]
    return np.asarray([bvx, bvy])

# Plot outline around line objects
class OutlineCollection(PatchCollection):
    """
    Source: @importanceofbeingernest
    https://stackoverflow.com/questions/55911075/how-to-plot-the-outline-of-the-outer-edges-on-a-matplotlib-line-in-python/56030879#56030879
    """
    def __init__(self, linecollection, ax=None, **kwargs):
        self.ax = ax or plt.gca()
        self.lc = linecollection
        assert np.all(np.array(self.lc.get_segments()).shape[1:] == np.array((2,2)))
        rect = plt.Rectangle((-.5, -.5), width=1, height=1)
        super().__init__((rect,), **kwargs)
        self.set_transform(mtrans.IdentityTransform())
        self.set_offsets(np.zeros((len(self.lc.get_segments()),2)))
        self.ax.add_collection(self)

    def draw(self, renderer):
        segs = self.lc.get_segments()
        n = len(segs)
        factor = 72/self.ax.figure.dpi
        lws = self.lc.get_linewidth()
        if len(lws) <= 1:
            lws = lws*np.ones(n)
        transforms = []
        for i, (lw, seg) in enumerate(zip(lws, segs)):
            X = self.lc.get_transform().transform(seg)
            mean = X.mean(axis=0)
            angle = np.arctan2(*np.squeeze(np.diff(X, axis=0))[::-1])
            length = np.sqrt(np.sum(np.diff(X, axis=0)**2))
            trans = mtrans.Affine2D().scale(length,lw/factor).rotate(angle).translate(*mean)
            transforms.append(trans.get_matrix())
        self._transforms = transforms
        super().draw(renderer)

# =====================
# NetworkX Wrappers
# =====================
def draw_networkx_labels_with_box(G, pos,
                         labels=None,
                         font_size=15,
                         font_color=None,
                         font_family='sans-serif',
                         font_weight='normal',
                         alpha=1.0,
                         node_box_facecolor=None,
                         node_box_edgecolor=None,
                         node_alpha=0.85,
                         boxstyle='round,pad=1',
                         linewidth=1.618,
                         ax=None,
                         **kwds):
    # https://github.com/networkx/networkx/blob/master/networkx/drawing/nx_pylab.py
    """
    __future___
    * Clean up the code for `d_nodeface_alpha`
    """
    if ax is None:
        ax = plt.gca()

    if labels is None:
        labels = list(G.nodes())

    if not is_dict_like(labels):
        labels = dict((n, n) for n in labels)

    # Set colors
    def _process_colors(G, DEFAULT, colors):
        if colors is None:
            colors = DEFAULT
        if is_color_like(colors):
            colors = dict(zip(G.nodes(), G.number_of_nodes()*[colors]))
        if isinstance(colors, pd.Series):
            colors = colors.to_dict()
        return {n:colors[n] if n in colors else DEFAULT for n in G.nodes()}

    # Default
    DEFAULT_font_color = "#000000"
    DEFAULT_node_box_facecolor = "#d8d8d8"
    DEFAULT_node_box_edgecolor =  "#000000"
    DEFAULT_node_box_alpha = 0.85


    # Colors
    d_nodefont_color = _process_colors(G, DEFAULT_font_color, font_color)
    d_nodeface_color = _process_colors(G, DEFAULT_node_box_facecolor, node_box_facecolor)
    d_nodeedge_color = _process_colors(G, DEFAULT_node_box_edgecolor, node_box_edgecolor)
    # Alpha
    if is_number(node_alpha):
        d_nodeface_alpha = {node:node_alpha for node in d_nodeface_color}
    else:
        assert is_dict_like(node_alpha), "If `node_alpha` is not a single alpha value then it must be dict-like of {node_label:alpha}"
        d_nodeface_alpha = dict()
        for n, label in labels.items():
            if label in node_alpha:
                d_nodeface_alpha[n] = node_alpha[n]
            else:
                d_nodeface_alpha[n] = DEFAULT_node_box_alpha


    # set optional alignment
    horizontalalignment = kwds.get('horizontalalignment', 'center')
    verticalalignment = kwds.get('verticalalignment', 'center')

    text_items = {}  # there is no text collection so we'll fake one
    for n, label in labels.items():
        if n in pos:
            (x, y) = pos[n]
            if not nx.utils.is_string_like(label):
                label = str(label)  # this makes "1" and 1 labeled the same
            t = ax.text(x, y,
                        label,
                        size=font_size,
                        color=d_nodefont_color[n],
                        family=font_family,
                        weight=font_weight,
                        alpha=alpha,
                        horizontalalignment=horizontalalignment,
                        verticalalignment=verticalalignment,
                        transform=ax.transData,
                        bbox=dict(facecolor=d_nodeface_color[n], edgecolor=d_nodeedge_color[n], boxstyle=boxstyle, alpha=d_nodeface_alpha[n], linewidth=linewidth),
                        clip_on=True,
                        )
            text_items[n] = t
    return text_items
# ===================
# Matplotlib/Seaborn/Matplotlib_venn Wrappers
# ===================
# Scatter plots
def plot_scatter(
                 # Main
                 data:pd.DataFrame,
                 x=None,
                 y=None,
                 z=None,
                 c="gray",
                 s=None,
                 # Labels
                 title=None,
                 xlabel=None,
                 ylabel=None,
                 zlabel=None,
                 cbar_label=None,
                 show_labels=None,
                 color_labels=None,
                 show_grid=False,
                 show_axis=True,

                 # Plotting params
                 figsize=(8,8),
                 ax=None,
                 edgecolor="white",
                 linewidth=1,
                 alpha=0.9,
                 logx=False,
                 logy=False,
                 logz=False,
                 legend=None,
                 annot=None,
                 annot_jitter=None,
                 cbar=True,
                 box_alpha=0.85,
                 auto_scale=True,

                 # Aspects
                 share_axis=False,
                 grid_granularity=None,
                 focus=None,
                 pad_focus="auto",
                 xlim=None,
                 ylim=None,
                 zlim=None,
                 cmap = None,
                 vmin=None,
                 vmax=None,
                 style="seaborn-white",
                 dimensions=2,

                 # Ellipse
                 ellipse_data=None,
                 ellipse_linewidth=3,
                 ellipse_linestyle="-",
                 ellipse_fillcolor="gray",
                 ellipse_edgecolor="black",
                 ellipse_n_std = 3,
                 ellipse_alpha=0.333,
                 ellipse_fill=False,

                 # 3D-Specific Parameters
                 elev_3d=None,
                 azim_3d=None,
                 auto_scale_3d=True,
                 pane_color_3d = None,

                 # Keywords
                 fig_kws=dict(),
                 annot_kws=dict(),
                 legend_kws=dict(),
                 cbar_kws=dict(),
                 axis_kws=dict(fontsize=15),
                 title_kws=dict(),
                 label_kws=dict(),

                 # Misc
                 pseudocount=1e-5,
                 missing_color='#FFFFF0',
                 missing_alpha=1.0,
                 verbose=True,
                 func_centroid=np.mean,
                 **args
                 ):
    """
    annot: [(text, (x,y))]
    Future:
        Fix the annotations so it's (x,y): annot_text and incorporate data.index (i.e. show_labels?)
        Add a polar coordinate conversion option
    09.26.2017
        Updated to clean up color mapping.  Force all alphas as None and if alpha is supplied then added within rgba
    10.02.2017
        Adding 3D support. Warning `lims` may not work.  Removing option for not having `data`.  Created `configure_scatter_utility`
        to do preprocessing.
    """
    # Keywords
    if figsize=="slides":
        figsize=(26.18,  14.7)
    _fig_kws = dict(figsize=figsize)
    _fig_kws.update(fig_kws)
    _legend_kws = {'bbox_to_anchor': (1, 0.5), 'edgecolor': 'black', 'facecolor': 'white', 'fontsize': 15, 'frameon': True, 'loc': 'center left'}
    _legend_kws.update(legend_kws)
    _cbar_kws = dict()
    _cbar_kws.update(cbar_kws)
    _title_kws = dict(fontsize=15,fontweight="bold")
    _title_kws.update(title_kws)
    _annot_kws = {**axis_kws,  "color":"black", "bbox":{"facecolor":"white", "edgecolor":"black", "alpha":box_alpha, "linewidth":2}, }
    _annot_kws.update(annot_kws)

    # ==================================================
    # Configure plot input data
    # ==================================================
    args_utility = dict(
                         data=data,
                         x=x,
                         y=y,
                         z=z,
                         c=c,
                         s=s,
                         alpha=alpha,
                         # Aspects
                         cmap = cmap,
                         vmin=vmin,
                         vmax=vmax,
                         dimensions=dimensions,
                         # Missing data
                         missing_color=missing_color,
                         missing_alpha=missing_alpha,
                         verbose=verbose,)

    configured_plot_data = configure_scatter(**args_utility)
    # Unpack configured data
    coords = configured_plot_data["coords"]
    c = np.asarray(configured_plot_data["colors"])
    s = np.asarray(configured_plot_data["sizes"])
    cmap = configured_plot_data["cmap"]
    scalar_vector = np.asarray(configured_plot_data["scalar_vector"])

    # Unpack Coordinates
    x = np.asarray(coords["x"])
    y = np.asarray(coords["y"])
    if dimensions == 3:
        z = np.asarray(coords["z"])

    # ==================================================
    # Log scale
    # ==================================================
    if logx:
        x = np.log(x + pseudocount)
        if xlabel:
            xlabel = "log(%s)"%xlabel
    if logy:
        y = np.log(y + pseudocount)
        if ylabel:
            ylabel = "log(%s)"%ylabel
    if logz:
        z = np.log(z + pseudocount)
        if zlabel:
            zlabel = "log(%s)"%zlabel

    with plt.style.context(style):
        # Matplotlib
        if ax is None:
            fig = plt.figure(**_fig_kws)
            if dimensions == 2:
                ax = plt.subplot(111, polar=False)
            if dimensions == 3:
                ax = fig.add_subplot(111, projection="3d")
        else:
            fig = plt.gcf()

        # ==================================================
        # Polar transformation
        # ==================================================
#         if (polar == True) and (polar_transform_x):
#             x = linear2circular(x)
        # ==================================================
        # Plot
        # ==================================================
        if dimensions == 2:
            im = ax.scatter(x=x, y=y, c=c, s=s, edgecolor=edgecolor, linewidth=linewidth, alpha=None, cmap=cmap, **args)

        if dimensions == 3:
            im = ax.scatter(xs=x, ys=y, zs=z, c=c, s=s, edgecolor=edgecolor, linewidth=linewidth, alpha=None, cmap=cmap, **args)

            # Change angle
            ax.view_init(elev=elev_3d, azim=azim_3d)
            # Change pane color
            if pane_color_3d:
                ax.w_xaxis.set_pane_color(pane_color_3d)
                ax.w_yaxis.set_pane_color(pane_color_3d)
                ax.w_zaxis.set_pane_color(pane_color_3d)

            # Auto scaling
            if auto_scale_3d:
                # X
                scaled_xlim_min = x.min() - x.min()*1e-3
                scaled_xlim_max = x.max() + x.min()*1e-3
                scaled_xlims = (scaled_xlim_min, scaled_xlim_max)
                # Y
                scaled_ylim_min = y.min() - y.min()*1e-3
                scaled_ylim_max = y.max() + y.min()*1e-3
                scaled_ylims = (scaled_ylim_min, scaled_ylim_max)

                # Z
                scaled_zlim_min = z.min() - z.min()*1e-3
                scaled_zlim_max = z.max() + z.min()*1e-3
                scaled_zlims = (scaled_zlim_min, scaled_zlim_max)

                # Set
                ax.set_xlim(scaled_xlims)
                ax.set_ylim(scaled_ylims)
                ax.set_zlim(scaled_zlims)

        # ==================================================
        # Labels
        # ==================================================
        if title:
            ax.set_title(str(title), **_title_kws)
        if xlabel:
            ax.set_xlabel(str(xlabel), **axis_kws)
        if ylabel:
            ax.set_ylabel(str(ylabel), **axis_kws)
        if zlabel:
            ax.set_zlabel(str(zlabel), **axis_kws)
        # ==================================================
        # Grid
        # ==================================================
        ax.grid(show_grid)
        if grid_granularity is not None:
            ax.set_xticks(np.linspace(*ax.get_xlim(), grid_granularity))
            ax.set_yticks(np.linspace(*ax.get_ylim(), grid_granularity))
        if show_axis == False:
            ax.set_axis_off()
        # ==================================================
        # Annotations
        # ==================================================
        if annot is not None:
            # Make this handle other colors
            if isinstance(annot, pd.Series):
                annot = annot.to_dict()
            if is_dict(annot):
                annot = list(annot.items())

            # Format the data into a list
            # NOTE! Currently cannot have duplicate indicies or {s:xy1, s:xy2}
            # ___________________________
            # Is data a list of strings to index?
            if not is_nonstring_iterable(annot[0]):
                annot = [(k, xy.as_matrix()) for k, xy in coords.loc[annot,:].T.iteritems()]
            # Is data a {s:[id_1, id_2, ..., id_m]} object?
                # Assumes the lists are greater than 2
            if is_nonstring_iterable(annot[0][1]) and (len(annot[0][1]) > 2):
                annot = get_coords_centroid(coords, annot, metric=func_centroid)

            for i,(annot_text, (x_coord,y_coord)) in enumerate(annot):
                xy = np.array((x_coord,y_coord))
                if annot_jitter:
                    xy += np.random.RandomState(i).normal(loc=0, scale=annot_jitter, size=1)
                ax.annotate(annot_text, xy, **_annot_kws)
        # ==================================================
        # Ellipses
        # ==================================================
        if ellipse_data is not None:
            # Format Data
            if not isinstance(ellipse_data, pd.DataFrame):
                ellipse_data = get_parameters_ellipse(coords, groups=ellipse_data, metric=func_centroid, n_std=ellipse_n_std)

            # Default Ellipse Attributes
            # Edge Color
            if (not isinstance(ellipse_edgecolor, pd.Series)) and (not is_dict(ellipse_edgecolor)):
                if not is_color(ellipse_edgecolor):
                    ellipse_edgecolor = "black"
                    if verbose:
                        print(f"{ellipse_edgecolor} is invalid.  Defaulting to `black`", file=sys.stderr)
            if is_color(ellipse_edgecolor):
                ellipse_edgecolor = pd.Series([ellipse_edgecolor]*ellipse_data.shape[0], index=ellipse_data.index, name="ellipse_edgecolor")
            if is_dict(ellipse_edgecolor):
                ellipse_edgecolor = pd.Series(ellipse_edgecolor, name="ellipse_edgecolor")
            # Fill Color
            if (not isinstance(ellipse_fillcolor, pd.Series)) and (not is_dict(ellipse_fillcolor)):
                if not is_color(ellipse_fillcolor):
                    ellipse_fillcolor = "black"
                    if verbose:
                        print(f"{ellipse_fillcolor} is invalid.  Defaulting to `black`", file=sys.stderr)
            if is_color(ellipse_fillcolor):
                ellipse_fillcolor = pd.Series([ellipse_fillcolor]*ellipse_data.shape[0], index=ellipse_data.index, name="ellipse_fillcolor")
            if is_dict(ellipse_fillcolor):
                ellipse_fillcolor = pd.Series(ellipse_fillcolor, name="ellipse_fillcolor")


            # Linestyle
            if type(ellipse_linestyle) == str:
                ellipse_linestyle = pd.Series([ellipse_linestyle]*ellipse_data.shape[0], index=ellipse_edgecolor.index, name="ellipse_linestyle")
            # Linewidth
            if is_number(ellipse_linewidth):
                ellipse_linewidth = pd.Series([ellipse_linewidth]*ellipse_data.shape[0], index=ellipse_edgecolor.index, name="ellipse_linewidth")
            for group, Se_data in ellipse_data.iterrows():
                ellipse_object = Ellipse(xy=Se_data[["x","y"]],
                                         width=Se_data["width"],
                                         height=Se_data["height"],
                                         angle=Se_data["theta"],
                                         fill=ellipse_fill,
                                         edgecolor=ellipse_edgecolor[group],
                                         facecolor=ellipse_fillcolor[group],

                                         linewidth=ellipse_linewidth[group],
                                         linestyle=ellipse_linestyle[group],
                                         alpha=ellipse_alpha,
                                            )
                ax.add_artist(ellipse_object)
        # ==================================================
        # Point labels
        # ==================================================
        if show_labels:
            # Coloring labels
            if is_dict(color_labels):
                color_labels = pd.Series(color_labels)
            # Default to black
            if color_labels is None:
                color_labels = "black"
            # Color all values same color
            if type(color_labels) == str:
                color_labels = pd.Series([color_labels]*len(x), index=data.index)
            # If True, set colors to match scatter points
            if not isinstance(color_labels, pd.Series):
                if color_labels == True:
                    color_labels = c.copy()
            # Default all the labels if True
            if type(show_labels) == bool:
                if show_labels == True:
                    show_labels = data.index
            # Plot Text
            for obsv_id in show_labels:
                if obsv_id in data.index:
                    if dimensions == 2:
                        ax.text(x=coords.loc[obsv_id, "x"],
                                y=coords.loc[obsv_id, "y"],
                                s=obsv_id,
                                color=color_labels[obsv_id],
                                **label_kws)
                    if dimensions == 3:
                        ax.text(x=coords.loc[obsv_id, "x"],
                                y=coords.loc[obsv_id, "y"],
                                z=coords.loc[obsv_id, "z"],
                            s=obsv_id,
                            color=color_labels[obsv_id],
                            **label_kws)
        # ==================================================
        # Sharing axis
        # ==================================================
        # Match axis scales
        if share_axis == True:
            lims = coords.as_matrix().ravel()
            min_lim, max_lim = min(lims), max(lims)
            ax.set_xlim((min_lim*0.98,max_lim*1.02))
            ax.set_ylim((min_lim*0.98,max_lim*1.02))
            if dimensions == 3:
                ax.set_zlim((min_lim*0.98,max_lim*1.02))

        # ==================================================
        # Legend
        # ==================================================
        if legend is not None:
            ax.legend(*format_mpl_legend_handles(legend), **_legend_kws)
        # ==================================================
        # Focus on an area
        # ==================================================
        if focus:
            # Not tested for 3D
            if type(focus) == str:
                focus = data.loc[focus,:]
            if pad_focus == "auto":
                sub = lambda x: x[0] - x[1]
                pad_focus = (sub(ax.get_xlim()) + sub(ax.get_ylim()))*0.01

            ax.set_xlim((focus[0] - pad_focus, focus[0] + pad_focus))
            ax.set_ylim((focus[1] - pad_focus, focus[1] + pad_focus))
        # ==================================================
        # Limits
        # ==================================================
        if xlim:
            ax.set_xlim(xlim)
        if ylim:
            ax.set_ylim(ylim)
        if dimensions == 3:
            if zlim:
                ax.set_zlim(zlim)
        # ==================================================
        # Autoscaling
        # ==================================================
        if auto_scale:
            ax.autoscale_view()
        # ==================================================
        # Colorbar
        # ==================================================
        if cbar:
            if cmap is not None:
                try:
                    add_cbar_from_data(fig=fig, cmap=cmap, data=scalar_vector, label=cbar_label, cbar_kws=cbar_kws, vmin=vmin, vmax=vmax)
                except AssertionError:
                    print("Warning: Colorbar was not added because `vmin` and `vmax` are Nonetype and are necessary when color values are not scalar", file=sys.stderr)
        return fig, ax, im

# Venn diagrams
def plot_venn_diagram(subsets, labels=None, colors=None, palette="Set1", edgecolor="white", linestyle="-", linewidth=1.618, fontsize_labels = 15, fontsize_numbers=15, alpha=0.8, style="seaborn-white", normalize_to=1.0, desat=None, figsize=(5,5), ax=None, circ_kws=dict(), title=None, title_kws=dict()):
    """
    return fig, ax, (venn, circ)
    """
    with plt.style.context(style):

        # Figure
        if ax is None:
            fig, ax = plt.subplots(figsize=figsize)
        else:
            fig = plt.gcf()

        # Keywords
        _circ_kws = dict(alpha=1.0, color=edgecolor, linestyle=linestyle, linewidth=linewidth, ax=ax)
        _circ_kws.update(circ_kws)
        _title_kws = dict(fontsize=15, fontweight="bold")
        _title_kws.update(title_kws)

        # Force set type
        subsets = [*map(set, subsets)]
        # Number of subsets
        n_subsets = len(subsets)

        # Set venn objects and defaults
        if n_subsets == 2:
            venn_obj= matplotlib_venn.venn2
            circ_obj = matplotlib_venn.venn2_circles
            if labels is None:
                labels = ("A","B")
        if n_subsets == 3:
            venn_obj = matplotlib_venn.venn3
            circ_obj = matplotlib_venn.venn3_circles
            if labels is None:
                labels = ("A","B","C")
        if n_subsets > 3:
            print(f"Too many subsets: {n_subsets}")
        labels = [*map(str, labels)]

        # Default colors
        if colors is None:
            colors = sns.color_palette(palette=palette, n_colors=n_subsets, desat=None)

        # Create venn diagram
        venn = venn_obj(subsets=subsets, set_labels=labels, set_colors=colors, ax=ax, alpha=alpha, normalize_to=normalize_to)
        circ = circ_obj(subsets=subsets, **_circ_kws)

        # Title
        if title:
            ax.set_title(title, **_title_kws)

        for text in venn.set_labels:
            if text is not None:
                text.set_fontsize(fontsize_labels)
                text.set_fontweight("bold")

        for text in venn.subset_labels:
            if text is not None:
                text.set_fontsize(fontsize_numbers)

        return fig, ax, (venn, circ)

# Plot waterfall-style bar chart
def plot_waterfall(data:pd.Series,
                   title=None,
                   color=None,
                   cmap=plt.cm.seismic_r,
                   linewidth=1.618,
                   alpha=1,
                   edgecolor="black",
                   show_labels=True,
                   metric_label=None,
                   observation_label=None,
                   style="seaborn-white",
                   engine="matplotlib",
                   figsize=None,
                   ax=None,
                   metric_tick_fontsize=15,
                   observation_tick_fontsize=15,
                   rotation="auto",
                   orientation="vertical",
                   ascending=None,
                   errorbars=None,
                   label_kws=dict(),
                   title_kws=dict(),
                   bar_kws=dict(),
                   error_kws=dict(),
                   vmin=None,
                   vmax=None,
                   divergent="infer",
                   midpoint=0,
                   check_null=True,
                  ):
    orientation = orientation.lower().strip()
    assert engine in {"matplotlib", "pandas"}, "`engine` must be either matplotlib or pandas"
    if check_null:
        assert data.isnull().sum() == 0, "Please remove null values and try again"
    # Orientation
    if figsize is None:
        figsize = {"horizontal":(10,6.18033), "vertical":(6.18033,10)}[orientation]
    if rotation == "auto":
        rotation = {"horizontal":90, "vertical":0}[orientation]
    if ascending is not None:
        data = data.sort_values(ascending=ascending)
    # Keywrods
    _bar_kws = {"alpha":alpha,"linewidth":linewidth, "edgecolor":edgecolor }
    _bar_kws.update(bar_kws)
    _error_kws = {"color":"black", "fmt":"none"}
    _error_kws.update(error_kws)
    _label_kws = {"fontsize":18}
    _label_kws.update(label_kws)

    # Limits
    if divergent == "infer":
        divergent = all([data.min() < 0, data.max() > 0])

    if divergent:
        if vmax is None:
            vmax = data.abs().max()
        if vmin is None:
            vmin = -vmax
    else:
        if vmax is None:
            vmax = data.max()
        if vmin is None:
            vmin = data.min()
    if color is None:
        color = scalarmapping_from_data(data, cmap=cmap, vmin=vmin, vmax=vmax, alpha=_bar_kws["alpha"], mode=1)

    # Data
    if errorbars is not None:
        errorbars = errorbars[data.index]
    x = np.arange(data.size)
    y = data.values

    # Plotting
    with plt.style.context(style):
        if ax is None:
            fig, ax = plt.subplots(figsize=figsize)
        else:
            fig = plt.gcf()
        if orientation == "horizontal":

            if engine == "pandas":
                data.plot(kind="bar",
                               color=color,
                               ax=ax,
                              **_bar_kws,
                         )
            if engine == "matplotlib":
                ax.bar(x, y, color=color, **_bar_kws)
                if midpoint is not None:
                    ax.axhline(midpoint,color="black")
            if errorbars is not None:
                ax.errorbar(x,y,yerr=errorbars, **_error_kws)

            _ = ax.set_xlim((-0.5, data.size-0.25))
            _ = ax.set_xticks(np.arange(data.size))
            _ = ax.set_xticklabels(data.index, rotation=rotation, fontsize=observation_tick_fontsize)
            _ = ax.set_yticklabels(ax.get_yticks(), fontsize=metric_tick_fontsize)
            if show_labels == False:
                ax.set_xticklabels([])
        if orientation == "vertical":
            if engine == "pandas":
                data.plot(kind="barh",
                               alpha=alpha,
                               color=color,
                               ax=ax)
            if engine == "matplotlib":
                ax.barh(x, y, color=color,  **_bar_kws)
            if errorbars is not None:
                ax.errorbar(y,x,xerr=errorbars, **_error_kws)
                if midpoint is not None:
                    ax.axvline(midpoint,color="black")
            _ = ax.set_ylim((-0.5, data.size-0.25))
            _ = ax.set_yticks(np.arange(data.size))
            _ = ax.set_yticklabels(data.index, rotation=rotation, fontsize=observation_tick_fontsize)
            _ = ax.set_xticklabels(ax.get_xticks(), fontsize=metric_tick_fontsize)
            if show_labels == False:
                ax.set_yticklabels([])

        # Labels
        if observation_label:
            if orientation == "vertical":
                ax.set_ylabel(observation_label, fontsize=18)
            if orientation == "horizontal":
                ax.set_xlabel(observation_label, fontsize=18)
        if metric_label:
            if orientation == "vertical":
                ax.set_xlabel(metric_label, fontsize=18)
            if orientation == "horizontal":
                ax.set_ylabel(metric_label, fontsize=18)
        if title:
            ax.set_title(str(title), fontsize=18, fontweight="bold")
        return fig,ax

# Adding annotations onto plots
def plot_annotation(labels, x:pd.Series, y:pd.Series, ax, adjust_label_positions=True,  x_pad=0, y_pad=0, **kwargs):
    """
    Dependency: https://github.com/Phlya/adjustText
    * Need to get the **kwargs and *args
    """
    def _get_text_objects(labels, x, y, ax, **kwargs):
        text_objects = []
        for ((id, label), x_i, y_i) in zip(labels.iteritems(),x,y):
            if is_nonstring_iterable(label):
                label = label.values[0]
            text_objects.append(ax.text(x_i, y_i, label,  **kwargs))
        return text_objects
    if not is_dict_like(labels):
        labels = dict(zip(labels, labels))
    labels = pd.Series(labels)
    n = labels.index.size
    assert n >= 1, "There are no labels"
    x = pd.Series(x) + x_pad
    y = pd.Series(y) + y_pad
    assert set(labels.index) <= set(x.index), "All `labels` must be in `x.index"
    assert set(labels.index) <= set(y.index), "All `labels` must be in `y.index"
    x = x[labels.index].values
    y = y[labels.index].values

    text_objects = _get_text_objects(labels, x, y, ax, **kwargs)
    if adjust_label_positions:
        adjust_text(text_objects, x, y, ax=ax)
    return text_objects

# Volcano plot
def plot_volcano(
    diffusion_values:pd.Series, 
    test_values:pd.Series, 
    tol_diffusion=0, 
    tol_test=0.05, 
    test_field="FDR", 
    xlabel="log$_2$(FC)", 
    ylabel="-log$_2$(FDR)", 
    horizontal=None, 
    linestyle=":", 
    size_significant=50, 
    size_nonsignificant=18, 
    ax=None, figsize=(5,5), 
    cmap=CMAP_DIVERGING, # This is from soothsayer.utils
    color_nonsignificant="#f8f8ff", 
    title=None, alpha_ratio=0.618, 
    linewidth=0.618,  
    edgecolor="black", 
    style="seaborn-white", 
    fill_nonsignificant="gray", 
    alpha_fill=0.1618, 
    show_annotations=False, 
    adjust_annotation_positions=True, 
    annot_kws=dict(),  
    *args):
    """
    # Generalized to all types of metrics (not just fold-change)
    # Added `test_field` (2018-May-11)
    Need to make sure this works correctly with MultiIndex
    Make a general use of tol_test so it makes sense to use tol_pvalue or something
    """
    with plt.style.context(style):
        if ax is None:
            fig, ax = plt.subplots(figsize=figsize)
        else:
            fig = plt.gcf()
        # Keywords
        _annot_kws = {}
        _annot_kws.update(annot_kws)

        # Vectors
        x = diffusion_values.copy()
        y = test_values.copy()
        assert np.all(x.index == y.index), "diffusion_values and test_values must have the same index"
        assert x.index.value_counts().max() == 1, "Cannot have duplicate values in the index"

        # DEGs
        significant_diffusion = x.abs() >= tol_diffusion
        significant_test = y < tol_test
        idx_degs = (significant_diffusion.astype(int) + significant_test.astype(int))[lambda z: z == 2].index
        vmax = x[idx_degs].abs().max()

        # Plot data
        c = map_colors(diffusion_values, mode=3, cmap=cmap, vmin=-vmax, vmax=vmax, format="hex")
        nonsignificant_labels = pd.Index(list(set(x.index) - set(idx_degs)))
        if color_nonsignificant is not None:
            mask = c.index.map(lambda label: label in nonsignificant_labels).values.astype(bool)
            c[mask] = color_nonsignificant
        s = pd.Series(x.index.map(lambda z:{True:50, False:18}[ z in idx_degs]), index=x.index)

        # Horizontal Lines

        if horizontal is None:
            horizontal = [tol_test]

        cond_A = hasattr(horizontal, "__len__") == True
        cond_B = type(linestyle) == str
        if all([cond_A, cond_B]):
            linestyle = [linestyle]*len(horizontal)

        # Alphas
        if len(horizontal) > 1:
            alphas = [1] + ((np.array([1]*(len(horizontal) - 1))*alpha_ratio).cumsum()).tolist()
        else:
            alphas = [1]

        # Synthesize Plot
        df_plot = pd.DataFrame([x,-np.log2(y.astype(float))], index=["x","y"]).T
        fig, ax, im = plot_scatter(data=df_plot, x="x", y="y", c=c, s=s, xlabel=xlabel, ylabel=ylabel, ax=ax, title=title, linewidth=linewidth, edgecolor=edgecolor, *args)

        # Synth Horizontal
        for i in range(len(horizontal)):
            ax.axhline(-np.log2(horizontal[i]), color="black", linestyle=linestyle[i], alpha=alphas[i])

        # Synth Vertical
        ax.axvline(-tol_diffusion, color="black", linestyle=linestyle[0])
        ax.axvline(tol_diffusion, color="black", linestyle=linestyle[0])

        # Fill nonsignficant area
        xlim = ax.get_xlim()
        ylim = ax.get_ylim()
        if fill_nonsignificant is not None:
            ax.fill_between([-tol_diffusion, tol_diffusion], *ylim, color=fill_nonsignificant, alpha=alpha_fill)
            ax.fill_betweenx( [min(ylim), -np.log2(tol_test)], *xlim, color=fill_nonsignificant, alpha=alpha_fill)
            ax.set(xlim=xlim, ylim=ylim)

        # Labels
        cond_C = show_annotations is not False
        cond_D = show_annotations is not None
        if all([cond_C, cond_D]):
            if show_annotations is True:
                show_annotations = idx_degs
            if  isinstance(show_annotations, (str, int)):
                show_annotations = [show_annotations]
            plot_annotation(labels=show_annotations, x=df_plot["x"], y=df_plot["y"], ax=ax, adjust_label_positions=adjust_annotation_positions, **_annot_kws)
        return fig, ax, im

# Plot prevalence of attributes in compositional data
def plot_prevalence(X:pd.DataFrame, color="teal", edgecolor="black", interval_type="closed", figsize=(13,5),  ax=None, style="seaborn-white", show_prevalence=[ 0.382, 0.5, 0.618, 1.0], show_ygrid=True, show_xgrid=False, show_legend=True, cmap=plt.cm.Dark2_r, attr_type="attr", obsv_type = "obsv", title=None, fill=True, title_kws=dict(), legend_kws=dict(), legend_title="Prevalence", legend_title_kws=dict(), n_xticks=25, yscale="linear", xlabel="auto", ylabel="auto", scale=False):

    _title_kws = {"fontsize":18, "fontweight":"bold"}
    _title_kws.update(title_kws)

    _legend_kws = { 'edgecolor': 'black', 'facecolor': 'white', 'fontsize': 15, 'frameon': True, "title_fontsize":15}
    _legend_kws.update(legend_kws)

    _legend_title_kws = {"size":15, "weight":"bold"}
    _legend_title_kws.update(legend_title_kws)


    with plt.style.context(style):

        if ax is  None:
            fig, ax = plt.subplots(figsize=figsize)
        else:
            fig = plt.gcf()
        # Dimensions
        n,m = X.shape

        # Prevalence
        x_prevalence = prevalence(X=X, interval_type=interval_type, name=None, scale=scale)

        # Plot
        ax.plot(x_prevalence.index, x_prevalence, color="gray")
        ax.scatter(x_prevalence.index, x_prevalence, color=color, edgecolor=edgecolor, linewidth=1, alpha=0.95)

        # Specific prevalence values
        if show_prevalence is not None:
            if any(map(lambda x: isinstance(x,float), show_prevalence)):
                show_prevalence = np.asarray(show_prevalence)
                assert all([np.all(show_prevalence <= 1), np.all(show_prevalence > 0)]), "If float values are provided then they represent ratios (i.e., Interval = (0,1] ])"
                show_prevalence = (show_prevalence * n).astype(int)

            colors = Chromatic.from_continuous(y=x_prevalence, cmap=cmap, vmin=0, vmax=max(show_prevalence))
            label_color = OrderedDict()
            for x_pos in sorted(show_prevalence):
                y_pos = x_prevalence.loc[x_pos]
                color = colors.obsv_colors[x_pos]
                # label = "N$_{%s}$ = %d\tN$_{%s}$ = %d"%(obsv_type, x_pos, attr_type, y_pos)
                label = "N$_{%s}$ = %d\tN$_{%s}$ = %d"%(obsv_type.replace("_","\_"), x_pos, attr_type.replace("_","\_"), y_pos) # 2019-04-19

                im = ax.plot([x_pos,x_pos], [0,y_pos], color=color, linestyle="-", linewidth=1, alpha=1, label=label)
                im = ax.plot([0 ,x_pos], [y_pos,y_pos], color=color, linestyle="-", linewidth=1, alpha=1)
                label_color[label] = color
                if fill:
                    im = ax.fill_between([0,x_pos], [y_pos,y_pos], color=color, alpha=0.1618)
            if show_legend:
                ax.legend(*format_mpl_legend_handles(label_color), **_legend_kws)
                if legend_title is not None:
                    ax.legend_.set_title(legend_title, prop=_legend_title_kws)

        # Set limits
        ax.set_ylim((0,x_prevalence.max()))
        ax.set_xlim((0,n))
        ax.set_xticks(np.linspace(0,n,min(n_xticks,n)).astype(int) )

        # Axis Labels
        if xlabel == "auto":
            if not scale:
                xlabel = f"Number of {obsv_type}s with $y_j$ {attr_type}s"
            else:
                xlabel = f"Ratio of {obsv_type}s with $y_j$ {attr_type}s"
        if ylabel == "auto":
            ylabel = f"Number of {attr_type}s with $x_i$ {obsv_type}s"
        ax.set_xlabel(xlabel, fontsize=15)
        ax.set_ylabel(ylabel, fontsize=15)

        # Show grids
        ax.xaxis.grid(show_xgrid)
        ax.yaxis.grid(show_ygrid)

        # Title
        if title is not None:
            ax.set_title(title, **_title_kws)
        return fig, ax

# Plot compositional data
def plot_compositional(
    X:pd.DataFrame,
    y:pd.Series=None,
    c:pd.Series=None,
    color_density="black",
    edgecolor="gray",
    s=None,
    class_colors=None,
    attr_type="attr",
    unit_type="read",
    figsize=(13,8),
    title=None,
    style="seaborn-white",
    palette="hls",
    vertical_lines = np.mean,
    horizontal_lines=np.mean,
    cmap=plt.cm.gist_heat_r,
    scatter_kws=dict(),
    show_xgrid=False,
    show_ygrid=True,
    show_density_2d=True,
    show_density_1d=True,
    show_legend=True,
    legend_kws=dict(),
    legend_title=None,
    title_kws=dict(),
    legend_title_kws=dict(),
    axis_label_kws=dict(),
    line_kws=dict(),
    kde_1d_kws=dict(),
    kde_2d_kws=dict(),
    panel_pad=0.1,
    panel_size=0.618,
    ax=None,
    shade_lowest_2d=False,
    split_class_density=True,
    background_color="white",
    show_annotations=False,
    adjust_annotation_positions=True,
    annot_kws=dict(),
    log_depth=True,
    log_richness=False,
    ):
    def _get_line_data(y, data, lines, class_colors, _line_kws):
        kws_collection = list()
        if is_function(lines):
            if class_colors is not None:
                placeholder = list()
                for id_class, idx_query in pd_series_collapse(y).iteritems():
                    v = lines(data[idx_query])
                    placeholder.append(v)
                    kws = {**_line_kws}
                    kws["color"] = class_colors[id_class]
                    kws_collection.append(kws)
                lines = placeholder
            else:
                lines = lines(data)
        if is_number(lines):
            lines = [lines]
        if len(kws_collection) == 0:
            kws_collection = [_line_kws]*len(lines)
        return lines, kws_collection



    # Data
    X = X.fillna(0)
    # RIchness
    richness = (X > 0).sum(axis=1)
    zeromask_richness = richness == 0
    if np.any(zeromask_richness):
        warnings.warn("Removing the following observations because richness = 0: {}".format(zeromask_richness.index[zeromask_richness]))
        richness = richness[~zeromask_richness]
    if log_richness:
        richness = np.log10(richness)
    # Depth
    depth = X.sum(axis=1)
    zeromask_depth = depth == 0
    if np.any(zeromask_depth):
        warnings.warn("Removing the following observations because depth = 0: {}".format(zeromask_depth.index[zeromask_depth]))
        depth = depth[~zeromask_depth]
    if log_depth:
        depth = np.log10(depth)
    # Overlap
    index = pd.Index(sorted(depth.index & richness.index))
    depth = depth[index]
    richness = richness[index]
    X = X.loc[index,:]

    # Defaults
    _title_kws = {"fontsize":15, "fontweight":"bold"}
    _title_kws.update(title_kws)
    _legend_kws = {'fontsize': 12, 'frameon': True, 'facecolor': 'white', 'edgecolor': 'black'}#, 'loc': 'center left', 'bbox_to_anchor': (1, 0.5)}
    _legend_kws.update(legend_kws)
    _legend_title_kws = {"size":15, "weight":"bold"}
    _legend_title_kws.update(legend_title_kws)
    _axis_label_kws = {"fontsize":15}
    _axis_label_kws.update(axis_label_kws)
    _kde_1d_kws = {"rug":max(X.shape) < 5000, "hist_kws":{"alpha":0.382}, "kde_kws":{"alpha":0.618}} # Rug takes too long when theres a lot points
    _kde_1d_kws.update(kde_1d_kws)
    _kde_2d_kws = {"shade":True, "shade_lowest":shade_lowest_2d, "alpha":0.618, "linewidth":1}
    _kde_2d_kws.update(kde_2d_kws)
    _line_kws = {"color":color_density, "linewidth":1.618, "linestyle":":", "alpha":1.0}
    _line_kws.update(line_kws)
    _annot_kws = {}
    _annot_kws.update(annot_kws)
    _scatter_kws={"s":s, "cmap":cmap, "vmin":0}
    _scatter_kws.update(scatter_kws)
    # Colors
    use_entropy = True
    if y is not None:
        assert set(y.index) >= set(X.index), "All elements of X.index must be in y.index"
        y = y[X.index]
        if class_colors is not None:
            obsv_colors = y.map(lambda id_obsv:class_colors[id_obsv])
        else:
            colors = Chromatic.from_classes(y, palette=palette)
            obsv_colors = colors.obsv_colors[index]
            class_colors = colors.class_colors
        use_entropy = False
    else:
        show_legend=False

    if use_entropy:
        c = X.apply(lambda x:entropy(x, base=2), axis=1)

    if c is not None:
        c = pd.Series(c)
        assert set(c.index) >= set(X.index), "Not all observations from `X` are in `c`"
        c = c[index]
        obsv_colors = c
        colors_global = c
    else:
        colors_global = color_density


    axes = list()
    with plt.style.context(style):
        # Set up axes
        if ax is None:
            fig, ax = plt.subplots(figsize=figsize)
        else:
            fig = plt.gcf()
        axes.append(ax)

        # Scatter plot
        if y is not None:
            for id_class, idx_query in pd_series_collapse(y).iteritems():
                ax.scatter(depth[idx_query], richness[idx_query], c=obsv_colors[idx_query], edgecolor=edgecolor, label=id_class, **_scatter_kws)
        else:
            ax.scatter(depth, richness, c=colors_global, edgecolor=edgecolor, **_scatter_kws)
        # Limits
        xlim = ax.get_xlim()
        ylim = ax.get_ylim()
        # Vertical lines
        if vertical_lines is not None:
            vertical_lines, kws_collection = _get_line_data(y, depth, vertical_lines, class_colors, _line_kws)
            for i, pos in enumerate(vertical_lines):
                ax.axvline(pos, **kws_collection[i])
        # Horizontal lines
        if horizontal_lines is not None:
            horizontal_lines, kws_collection = _get_line_data(y, richness, horizontal_lines, class_colors, _line_kws)
            for i, pos in enumerate(horizontal_lines):
                ax.axhline(pos, **kws_collection[i])
        ax.set(xlim=xlim, ylim=ylim)
        # Density (1D)
        if show_density_1d:
            # Setup
            divider = make_axes_locatable(ax)
            # Richness
            ax_right = divider.append_axes("right", pad=panel_pad, size=panel_size)
            ax_top = divider.append_axes("top", pad=panel_pad, size=panel_size)

            if split_class_density:
                if class_colors is None:
                    sns.distplot(richness, color=color_density, vertical=True, ax=ax_right, **_kde_1d_kws)
                    sns.distplot(depth, color=color_density, vertical=False, ax=ax_top, **_kde_1d_kws)
                else:
                    for id_class, idx_query in pd_series_collapse(y).iteritems():
                        sns.distplot(richness[idx_query], color=class_colors[id_class], vertical=True, ax=ax_right, **_kde_1d_kws)
                        sns.distplot(depth[idx_query], color=class_colors[id_class], vertical=False, ax=ax_top, **_kde_1d_kws)
            else:
                sns.distplot(richness, color=color_density, vertical=True, ax=ax_right, **_kde_1d_kws)
                sns.distplot(depth, color=color_density, vertical=False, ax=ax_top, **_kde_1d_kws)
            ax_right.set(ylim=ylim, xticklabels=[],yticklabels=[],yticks=ax.get_yticks())
            ax_top.set(xlim=xlim, xticklabels=[],yticklabels=[],xticks=ax.get_xticks())
            axes.append(ax_right)
            axes.append(ax_top)

        # Density (2D)
        if show_density_2d:
            if split_class_density:
                if class_colors is None:
                    sns.kdeplot(data=depth, data2=richness, color=color_density, zorder=0,  ax=ax, **_kde_2d_kws)
                else:
                    for id_class, idx_query in pd_series_collapse(y).iteritems():
                        try:
                            sns.kdeplot(data=depth[idx_query], data2=richness[idx_query],  color=class_colors[id_class], zorder=0, ax=ax, **_kde_2d_kws)
                        except Exception as e:
                            warnings.warn("({}) Could not compute the 2-dimensional KDE plot for the following class: {}".format(id_class))
            else:
                sns.kdeplot(data=depth, data2=richness, color=color_density, zorder=0,  ax=ax, **_kde_2d_kws)

        # Annotations
        conditions = [
            show_annotations is not False,
            show_annotations is not None
        ]
        if all(conditions):
            if show_annotations is True:
                show_annotations = X.index
            if  isinstance(show_annotations, (str, int)):
                show_annotations = [show_annotations]
            plot_annotation(labels=show_annotations, x=depth[show_annotations], y=richness[show_annotations], ax=ax, adjust_label_positions=adjust_annotation_positions, **_annot_kws)

        # Labels
        xlabel = "Depth [$N_{%s}$]"%(f"{unit_type}s")
        if log_depth:
            xlabel = "log$_{10}$(%s)"%(xlabel)
        ylabel = "Richness [$N_{%s}$]"%(f"{attr_type}s")
        if log_richness:
            ylabel = "log$_{10}$(%s)"%(ylabel)

        ax.set_xlabel(xlabel, **_axis_label_kws)
        ax.set_ylabel(ylabel, **_axis_label_kws)
        ax.xaxis.grid(show_xgrid)
        ax.yaxis.grid(show_ygrid)

        # Legend
        if show_legend:
            ax.legend(**_legend_kws)
            if bool(legend_title):
                ax.legend_.set_title(legend_title, prop=_legend_title_kws)
        # Title
        if title is not None:
            axes[-1].set_title(title, **_title_kws)

        # Background color
        for ax_query in axes:
            ax_query.set_facecolor(background_color)

        return fig, axes

# ===================
# Plot.ly Wrappers
# ===================
# ===================
# Bokeh Wrappers
# ===================



#!/usr/bin/env python
# =============================
# Creator: Josh L. Espinoza (J. Craig Venter Institute)
# Date[0]:Init: 2018-April-02
# Date[-1]:Current: 2018-August-28
# =============================
# BSD License
# =============================
# Copyright 2018 Josh L. Espinoza
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# Version
__version_clairvoyance__ = "v0.3_2018-08-28"

# Built-ins
import os, sys, multiprocessing, argparse, time, shutil, warnings
from ast import literal_eval
# warnings.simplefilter(action='ignore', category=FutureWarning) # FutureWarning: Series.compress(condition) is deprecated. Use 'Series[condition]' or 'np.asarray(series).compress(condition)' instead.
# warnings.simplefilter(action='ignore', category=UserWarning) # UserWarning: Attempting to set identical left==right results in singular transformations; automatically expanding
warnings.filterwarnings("ignore")
os.environ['KMP_DUPLICATE_LIB_OK']='True' # In case you have 2 versions of OMP installed

# PyData
import pandas as pd
import numpy as np

## Plotting
from matplotlib import use as mpl_backend
mpl_backend("Agg")
import matplotlib.pyplot as plt

# Machine learning
from sklearn import model_selection
from sklearn.linear_model import LogisticRegression
from sklearn.tree import DecisionTreeClassifier
# from sklearn.multiclass import OneVsRestClassifier
from sklearn.base import clone

# Utility
from tqdm import tqdm, trange

# Soothsayer
from soothsayer.utils import format_duration, create_logfile, determine_mode_for_logfiles, to_precision
from soothsayer.io import read_dataframe, read_object, write_object
from soothsayer.transmute.normalization import normalize
from soothsayer.feature_extraction import Clairvoyant

# Config
pd.set_option('max_colwidth',300)

# Functions
legend_kws = dict(fontsize=10, frameon=True, facecolor="white", edgecolor="black", loc='center left', bbox_to_anchor=(1, 0.5))

# Controller
def main(argv=None):
    parser = argparse.ArgumentParser(
    """
    soothsayer:clairvoyance {}
    """.format(__version_clairvoyance__)
    )
    # Input
    parser.add_argument('-X', '--attribute_matrix', type=str, help = 'Input: Path/to/Tab-separated-value.tsv of attribute matrix (rows=samples, cols=attributes)')
    parser.add_argument('-y', '--target_vector', type=str, help = 'Input: Path/to/Tab-separated-value.tsv of target vector (rows=samples, column=integer target)')
    parser.add_argument("-e", "--encoding", default=None, help="Input: Path/to/Tab-separated-value.tsv of encoding.  Column 1 has enocded integer and Column 2 has string representation.  No header! [Default: None]")

    # Normalization
    parser.add_argument("--normalize", type=str, default="None", help="Normalize the attribute matrix.  Valid Arguments: {'tss', 'zscore', 'quantile', None}: Warning, apply with intelligence.  For example, don't use total-sum-scaling for fold-change values. [Experimental: 2018-May-03]")

    # Hyperparameters
    parser.add_argument("-m", "--model_type", type=str, default="logistic,tree", help="Model types in ['logistic','tree'] represented by a comma-separated list [Default: 'logistic,tree']")
    parser.add_argument("--n_iter", type=int, default=10, help="Number of iterations to split data and fit models. [Default: 10]")
    parser.add_argument("--random_state", type=int, default=0, help="[Default: 0]")
    parser.add_argument("--n_jobs", type=int, default=1, help="[Default: 1]")
    parser.add_argument("--min_threshold", type=str, default="0.0,None", help="Minimum accuracy for models when considering coeficients for weights. Can be a single float or list of floats separated by `,` ranging from 0-1. If None it takes an aggregate weight.  [Default: 0.0,None]")
    parser.add_argument("--random_mode", type=int, default=0, help="Random mode 0,1, or 2.  0 uses the same random_state each iteration.  1 uses a different random_state for each parameter set.  2 uses a different random_state each time. [Default: 0]")
    parser.add_argument("--percentiles", type=str, default="0.0,0.3,0.5,0.75,0.9,0.95,0.96,0.97,0.98,0.99", help="Iterative mode takes in a comma-separated list of floats from 0-1 that dictate how many of the attributes will be used [Default: 0.0,0.3,0.5,0.75,0.9,0.95,0.96,0.97,0.98,0.99]")

    # Labels & Directories
    parser.add_argument("-n", "--name", type=str, default=time.strftime("%Y-%m-%d"), help="Name for data [Default: Date]")
    parser.add_argument("-o", "--out_dir", type=str, default = os.getcwd(), help = "Output: Path/to/existing-directory-for-output.  Do not use `~` to specify /home/[user]/ [Default: cwd]" )
    parser.add_argument("--attr_type", type=str, default="attr", help="Name of attribute type (e.g. gene, orf, otu, etc.) [Default: 'attr']")
    parser.add_argument("--class_type", type=str, default="class", help="Name of class type (e.g. phenotype, MOA, etc.) [Default: 'class']")

    # Cross Validation
    parser.add_argument("--method", type=str, default="bruteforce", help = "{'adaptive', 'bruteforce'} [Default: 'bruteforce']")
    parser.add_argument("--adaptive_range", type=str, default="0.0,0.5,1.0", help="Adaptive range in the form of a string of floats separated by commas [Default:'0.0,0.5,1.0']")
    parser.add_argument("--adaptive_steps", type=str, default="1,10", help="Adaptive stepsize in the form of a string of ints separated by commas. len(adaptive_range) - 1 [Default:'1,10']")
    parser.add_argument("--cv", default=5, type=str, help="Number of steps for percentiles when cross validating and dropping attributes.  If cv=0 then it skips cross-validation step. [Default: 5]")
    parser.add_argument("--min_bruteforce", type=int, default=150, help="Minimum number of attributes to adjust to bruteforce. If value is 0 then it will not change method [Default: 150]")
    parser.add_argument("--early_stopping", type=int, default=100, help="Stopping the algorthm if a certain number of iterations do not increase the accuracy. Use -1 if you don't want to use `early_stopping` [Default: 100]")


    # Reading in dataframes
    parser.add_argument("--compression", type=str, default="infer")
    parser.add_argument("--pickled", type=str, default="infer")

    # I/O Model
    parser.add_argument("--save_kernel", default=True, help="Save the kernel [Default: True]")
    parser.add_argument("--save_model", default=False, help="Save the model [Default: False]")
    parser.add_argument("--save_data", default=True, help="Save the model [Default: True]")


    # Broken parameters
    # parser.add_argument("--scoring", type=str, default="accuracy", help="Scoring metric [Default: 'accuracy'] NOTE ONLY USE ACCURACY FOR NOW")
    # parser.add_argument("--load_model", default="infer", help="path/to/model.pbz2.  Default: 'infer' which looks for model in the working directory")
    # parser.add_argument("--force_overwrite", default=False, help="Force overwrite data [Default: False]") # Note to fix this, all percentiles above the first change need to be recomputed



    opts = parser.parse_args(argv)

    opts.force_overwrite = True

    # opts.force_overwrite = True
    start_time = time.time()



    # Create working directory
    os.makedirs(opts.out_dir, exist_ok=True)
    path_synopsis = "{}/{}__synopsis".format(opts.out_dir,opts.name)
    os.makedirs(path_synopsis, exist_ok=True)
    log_info = create_logfile("information",
                                  "{}/{}.log".format(opts.out_dir,opts.name),
                                  mode=determine_mode_for_logfiles("{}/{}.log".format(opts.out_dir,opts.name), opts.force_overwrite)
    )

    # Initiating variables
    # ====================
    # Default Random State
    np.random.seed(opts.random_state)
    # Parallel
    if opts.n_jobs == -1:
        opts.n_jobs = multiprocessing.cpu_count()

    # File Paths
    if opts.out_dir.endswith("/"):
        opts.out_dir = opts.out_dir[:-1]

    # Adaptive ranges
    if opts.method == "adaptive":
        opts.adaptive_range = list(map(lambda x:float(x.strip()), opts.adaptive_range.split(",")))
        opts.adaptive_steps = list(map(lambda x:int(x.strip()), opts.adaptive_steps.split(",")))
        assert np.all(np.asarray(opts.adaptive_range) <= 1.0), "All values in `adaptive_range` should be: 0.0 <= x <= 1.0"
    # Early stopping
    if opts.early_stopping == -1:
        opts.early_stopping = np.inf

    # Minimum threshold
    if "," in opts.min_threshold:
        opts.min_threshold = opts.min_threshold.lower()
        special_values = list()
        if "none" in opts.min_threshold:
            opts.min_threshold = opts.min_threshold.replace("none","")
            special_values.append(None)
        opts.min_threshold = sorted(map(float, filter(bool, opts.min_threshold.split(",")))) + special_values
    else:
        if opts.min_threshold == "none":
            opts.min_threshold = [None]
        else:
            opts.min_threshold = [float(opts.min_threshold)]

    # Percentiles
    if "," in opts.percentiles:
        opts.percentiles = sorted(map(float, filter(bool, opts.percentiles.split(","))))
    else:
        if opts.percentiles.lower() in ["none", "false"]:
            opts.percentiles = [0.0]
        else:
            opts.percentiles = [float(opts.percentiles)]
    if 0.0 not in opts.percentiles:
        opts.percentiles = [0.0] + opts.percentiles
    assert all(map(lambda x: 0.0 <= x < 1.0, opts.percentiles)), "All percentiles must meet the following criteria: 0.0 <= x < 1.0"

    # Model types
    if "," in opts.model_type:
        opts.model_type = list(filter(bool, opts.model_type.split(",")))
    else:
        opts.model_type = [opts.model_type]
    assert set(opts.model_type) <= set(["logistic", "tree"]), "model_type should only be a combination of ['logistic','tree']. Current input: {}".format(opts.model_type)


    # Normalization
    if opts.normalize.lower() in ["false","none"]:
        opts.normalize = None

    # Serialization
    if opts.pickled != "infer":
        if opts.pickled.lower() in ["false", "none"]:
            opts.pickled = False
        if opts.pickled.lower() == "true":
            opts.pickled = True
    # Cross Validation Configure
    cv_repr = opts.cv
    if str(opts.cv).isnumeric():
        opts.cv = int(opts.cv)
        cv_labels = list(map(lambda x: "cv={}".format(x+1), range(opts.cv)))
    else:
        with open(opts.cv,"r") as f:
            _tmp_ncols_cv = len(f.read().split("\n")[0].split("\t"))
        # Is it just 2 columns (train, test) with no header?
        if  _tmp_ncols_cv == 2:
            opts.cv = pd.read_csv(opts.cv, header=None, sep="\t", dtype=str).applymap(literal_eval).values.tolist()
            cv_labels = list(map(lambda x: "cv={}".format(x+1), range(len(opts.cv))))
        # Are there 3 columns with a header with the first column being the name of cross-validation set?
        if _tmp_ncols_cv == 3:
            opts.cv = pd.read_csv(opts.cv, sep="\t", index_col=0, dtype=str).applymap(literal_eval).loc[:,["Training", "Testing"]]
            cv_labels = opts.cv.index
            opts.cv = opts.cv.values.tolist()

    # Saving model
    if type(opts.save_model) == str:
        if opts.save_model.lower() in ["false", "none", "no"]:
            opts.save_model = None
        else:
            opts.save_model = True

    # Saving model data
    if type(opts.save_data) == str:
        if opts.save_data.lower() in ["false", "none", "no"]:
            opts.save_data = None
        else:
            opts.save_data = True

    # Static data
    y = read_dataframe(opts.target_vector, compression=opts.compression, pickled=opts.pickled, func_index=str, func_columns=str, verbose=False)
    if isinstance(y, pd.Series):
        y = y.to_frame()
    y = y.iloc[:,0]

    # Encoding
    if opts.encoding:
        opts.encoding = pd.read_table(opts.encoding, sep="\t", header=None, index_col=0).iloc[:,0].to_dict()

    # Iterate through model types
    best_configuration = None
    baseline_scores_ = dict()
    summary_table = list()
    for m in opts.model_type:
        print("= == === ===== ======= ============ =====================", file=sys.stderr)
        print(m, file=sys.stderr)
        print("= == === ===== ======= ============ =====================".replace("=","."), file=sys.stderr)


        estimator = {"logistic":LogisticRegression(random_state=opts.random_state, multi_class="ovr", solver='liblinear'), "tree":DecisionTreeClassifier(random_state=opts.random_state)}[m]
        # Placeholders for dynamic data
        path_attributes_current = opts.attribute_matrix
        best_configuration_for_modeltype = None # (percentile, min_threshold, hyperparameters)
        best_accuracy_for_modeltype = 0
        best_hyperparameters_for_modeltype = None
        model = None
        best_attributes_for_modeltype = None

        log_crossvalidation = create_logfile(m,
                                                 "{}/{}__cross-validation.log".format(path_synopsis,m),
                                                 mode=determine_mode_for_logfiles("{}/{}__cross-validation.log".format(path_synopsis,m), opts.force_overwrite)
        )


        # Iterate through percentiles of weighted attributes
        baseline_percentile_0 = None
        for current_iteration, i_percentile in enumerate(opts.percentiles, start=0):

            # Determine whether or not to continue the analysis
            continue_analysis = True
            path_summary = "{}/{}/percentile_{}/summary.txt".format(opts.out_dir,m,i_percentile)
            if opts.force_overwrite == False:
                if os.path.exists(path_summary):
                    if os.path.getsize(path_summary) > 0:
                        continue_analysis = False
            # ==================================================================
            # Load previous analysis
            # ==================================================================
            # Check if directory exists, make one if not
            log_info.info("\t\t{} | Percentile={} | Creating temporary directories in {}".format(m,i_percentile,opts.out_dir))
            os.makedirs("{}/{}/percentile_{}".format(opts.out_dir,m,i_percentile), exist_ok=True)

            f_summary = open(path_summary, "w")
            # Load attributes
            log_info.info("\t\t{} | Percentile={} | Loading attributes datasets".format(m,i_percentile ))
            X = read_dataframe(path_attributes_current, compression=opts.compression, pickled=opts.pickled,  func_index=str, func_columns=str, verbose=False)
            if X.shape[1] < 2:
                print("Skipping current iteration {} because there is/are only {} attribute[s]".format(i_percentile, X.shape[1]), file=sys.stderr)
            else:
                # Normalize
                if opts.normalize is not None:
                    X = normalize(X, method=opts.normalize.lower(), axis=1) # COULD PROBABLY MOVE THIS UP AND COMBINE

                # Common Identifiers
                try:
                    assert np.all([X.index == y.index]), "X.index should be the same ordering as y.index"
                except ValueError:
                    A = set(X.index)
                    B = set(y.index)

                    print("set(X) - set(y) : len(A - B) ==>  {}".format(A - B),
                          "set(y) - set(X) : len(B - A) ==>  {}".format(B - A),
                          sep=2*"\n",
                          file=sys.stderr
                    )
                    sys.exit(1)



                params = pd.Series(opts.__dict__)
                if not continue_analysis:
                    log_info.info("\t\t{} | Percentile={} | Skipping analysis and loading data from: {}/{}/percentile_{}".format(m,i_percentile,opts.out_dir,m,i_percentile))

                    # Load model
                    model = read_object(path="{}/{}/percentile_{}/model.pbz2".format(opts.out_dir,m,i_percentile), compression="bz2")

                    # Set parameters for estimator and calculate baseline score
                    estimator.set_params(**model.best_hyperparameters_)
                    baseline_score = model_selection.cross_val_score(estimator, X=X.values, y=y.values, cv=opts.cv).mean()
                    baseline_scores_[(m,i_percentile)] = baseline_score

                    # Set original attribute size and baseline score
                    if i_percentile == 0.0:
                        n_attributes = X.shape[1]
                        baseline_percentile_0 = baseline_score
                    print("{} | Percentile={} | X.shape = {} | Baseline score({}) = {}".format(m, i_percentile, X.shape,path_attributes_current,to_precision(baseline_score)), file=sys.stderr)


                    # Determine if weights are different between any of the other runs
                    tol_difference = 1
                    if len(opts.min_threshold) > 1:
                        _current_order = np.asarray(model.extract_weights(min_threshold=opts.min_threshold[0], ascending=False).index)
                        for x in opts.min_threshold[1:]:
                            _query_order = np.asarray(model.extract_weights(min_threshold=x, ascending=False).index)
                            num_differences = np.sum(_current_order != _query_order)
                            if num_differences < tol_difference*2:  # Times 2 because differences come in pairs
                                opts.min_threshold.remove(x)
                                log_info.info("\t\t{} | Percentile={} | Removing min_accuracy={} from computation because {} ordering is already present from another threshold".format(m,i_percentile,x,model.attr_type))
                            else:
                                _current_order = _query_order

                    # Placeholders
                    best_accuracy_for_percentile = 0
                    best_min_threshold_for_percentile = None
                    best_weights_for_percentile = None
                    # best_results_for_percentile = None

                    for x in opts.min_threshold:
                        path_query = "{}/{}/percentile_{}/min_threshold_{}".format(opts.out_dir,m,i_percentile,x)
                        w_ = model.extract_weights(min_threshold=x, name=opts.name, ascending=False)

                        # Are there weights for this threshold?  If there's not then that means there are no models that predicted with this threshold or higher
                        if not w_.isnull().any():
                            scores_ = read_dataframe("{}/scores.tsv.gz".format(path_query), sep="\t", compression="gzip", verbose=False)

                            # Update the scores
                            idx_best_accuracy_for_minthreshold = scores_["accuracy"].sort_values(ascending=False).index[0]
                            query_accuracy = scores_.loc[idx_best_accuracy_for_minthreshold,"accuracy"]
                            query_sem = scores_.loc[idx_best_accuracy_for_minthreshold,"sem"]
                            attribute_set = scores_.loc[idx_best_accuracy_for_minthreshold, "{}_set".format(opts.attr_type)]

                            # Update accuracies for current percentile
                            if query_accuracy > best_accuracy_for_percentile:
                                best_min_threshold_for_percentile = x
                                best_accuracy_for_percentile = query_accuracy
                                best_weights_for_percentile = w_.copy()

                                # Update accuracies for current model
                                if query_accuracy > best_accuracy_for_modeltype:
                                    best_configuration_for_modeltype = (i_percentile, x)
                                    best_hyperparameters_for_modeltype = model.best_hyperparameters_
                                    best_accuracy_for_modeltype = query_accuracy
                                    best_attributes_for_modeltype = attribute_set

                            # Synthesize summary table
                            Se_query = pd.Series([m, model.best_hyperparameters_, i_percentile, x, baseline_percentile_0, baseline_score, query_accuracy,  query_sem, query_accuracy - baseline_percentile_0, opts.random_state, len(attribute_set), attribute_set],
                                           index=["model_type", "hyperparameters", "percentile",  "min_threshold", "baseline_score", "baseline_score_of_current_percentile", "accuracy", "sem", "delta", "random_state", "num_{}_included".format(opts.attr_type), "{}_set".format(opts.attr_type)])
                            summary_table.append(Se_query)
                # ==================================================================
                # Continue analysis
                # ==================================================================
                else:
                    if X.shape[1] <= opts.min_bruteforce:
                        method_crossvalidation = "bruteforce"
                        if opts.method is not "bruteforce":
                            log_info.info("\t\t{} | Percentile={} | Adjusting `method` to `bruteforce` because X.shape[1] <= {}".format(m,i_percentile,opts.min_bruteforce))
                    else:
                        method_crossvalidation = opts.method
                    # Run model
                    model = Clairvoyant(
                                model_type=m,
                                n_iter=opts.n_iter,
                                map_encoding=opts.encoding,
                                attr_type=opts.attr_type,
                                class_type=opts.class_type,
                                n_jobs=opts.n_jobs,
                                verbose=True,
                                random_state=opts.random_state,
                                random_mode=opts.random_mode,
                    )
                    log_info.info("\t\tFitting data")
                    model.fit(X, y, desc="{} | Percentile={} | Permuting samples and fitting models".format(m, i_percentile))
                    log_info.info("\t\t{} | Percentile={} | Best hyperparameters from fitting: {}".format(m,i_percentile,model.best_hyperparameters_))
                    log_info.info("\t\t{} | Percentile={} | Saving model: {}/{}/percentile_{}/model.pbz2".format(m,i_percentile,opts.out_dir,m,i_percentile))
                    if opts.save_model:
                        model.save_model(path="{}/{}/percentile_{}/model.pbz2".format(opts.out_dir,m,i_percentile), compression="bz2")


                    # Get kernel and weights
                    kernel_ = model.extract_kernel(into="xarray", name=opts.name)
                    acu_ = model.extract_accuracies(into="pandas", name=opts.name)
                    hyperparameters_ = model.extract_hyperparameters()
                    W_ = model.extract_weights(min_threshold=None, mode=1) # Future: Incldue any other min_threshold values that aren't in the linspace

                    log_info.info("\t\t{} | Percentile={} | Calculating weights".format(m,i_percentile))

                    # Set parameters for estimator and calculate baseline score
                    estimator.set_params(**model.best_hyperparameters_)
                    baseline_score = model_selection.cross_val_score(estimator, X=X.values, y=y.values, cv=opts.cv).mean()
                    baseline_scores_[(m,i_percentile)] = baseline_score

                    # Set original attribute size and baseline score
                    if i_percentile == 0.0:
                        n_attributes = X.shape[1]
                        baseline_percentile_0 = baseline_score
                    print("{} | Percentile={} | X.shape = {} | Baseline score({}) = {}".format(m,i_percentile,X.shape,path_attributes_current,to_precision(baseline_score)), file=sys.stderr)

                    acu_.to_frame("Accuracy").to_csv("{}/{}/percentile_{}/acu.tsv.gz".format(opts.out_dir,m,i_percentile), sep="\t", compression="gzip")
                    hyperparameters_.to_csv("{}/{}/percentile_{}/hyperparameters.tsv.gz".format(opts.out_dir,m,i_percentile), sep="\t", compression="gzip")
                    W_.to_csv("{}/{}/percentile_{}/weights.tsv.gz".format(opts.out_dir,m,i_percentile ), sep="\t", compression="gzip")



                    if opts.save_kernel:
                        kernel_.to_netcdf("{}/{}/percentile_{}/kernel.nc".format(opts.out_dir, m, i_percentile))
                    # Determine if weights are different
                    tol_difference = 1
                    if len(opts.min_threshold) > 1:
                        _current_order = np.asarray(model.extract_weights(min_threshold=opts.min_threshold[0], ascending=False).index)
                        for x in opts.min_threshold[1:]:
                            _query_order = np.asarray(model.extract_weights(min_threshold=x, ascending=False).index)
                            num_differences = np.sum(_current_order != _query_order)
                            if num_differences < tol_difference*2:  # Times 2 because differences come in pairs
                                opts.min_threshold.remove(x)
                                log_info.info("\t\t{} | Percentile={} | Removing min_accuracy={} from computation because {} ordering is already present from another threshold".format(m,i_percentile,x,model.attr_type))
                            else:
                                _current_order = _query_order
                    # Soothsayer | Attribute Finder
                    print("================================\nsoothsayer:clairvoyance {}\n================================".format(__version_clairvoyance__), file=f_summary)
                    print("Name:", opts.name, sep="\t", file=f_summary)
                    print("X:", path_attributes_current, sep="\t", file=f_summary)
                    print("y:", opts.target_vector, sep="\t", file=f_summary)
                    print("Encoding:", opts.encoding, sep="\t", file=f_summary)
                    print("Path:", opts.out_dir, sep="\t", file=f_summary)
                    print("Shape:", X.shape, sep="\t", file=f_summary)
                    print("Model Type:", m, sep="\t", file=f_summary)
                    print("Percentile:", i_percentile, sep="\t", file=f_summary)
                    # if opts.load_model:
                    #     print("Loaded Model:", opts.load_model, sep="\t", file=f_summary)
                    print("================\n Baseline \n================", file=f_summary)

                    print("Percentile = 0: ", baseline_percentile_0, file=f_summary)
                    print("Current percentile = {}:".format(i_percentile), baseline_score, file=f_summary)


                    print("================\n Hyperparameters \n================", file=f_summary)
                    print("\n".join(str(params[["n_iter", "random_state", "random_mode", "n_jobs", "min_threshold"]]).split("\n")[:-1]), file=f_summary)
                    print("================\n Labels \n================", file=f_summary)
                    print("\n".join(str(params[["attr_type", "class_type"]]).split("\n")[:-1]), file=f_summary)
                    print("================\n Cross Validation \n================", file=f_summary)
                    if opts.method == "adaptive":
                        print("\n".join(str(params[["method","adaptive_range","adaptive_steps", "early_stopping"]]).split("\n")[:-1]), file=f_summary)
                    if opts.method == "bruteforce":
                        print("\n".join(str(params[["method", "min_bruteforce", "early_stopping"]]).split("\n")[:-1]), file=f_summary)


                    print("cv", cv_repr, sep="\t", file=f_summary)
                    print("================\n Data Type \n================", file=f_summary)
                    print("\n".join(str(params[["compression", "pickled"]]).split("\n")[:-1]), file=f_summary)

                    # Placeholders
                    best_accuracy_for_percentile = 0
                    best_min_threshold_for_percentile = None
                    best_results_for_percentile = None
                    best_weights_for_percentile = None
                    # Run crossvalidation
                    run_crossvalidation = opts.cv is not 0
                    if run_crossvalidation:
                        # Cross Validation
                        for x in opts.min_threshold:
                            # Default to run the analysis.  First check if the path exists if does say if you want to overwrite.  If not then make the directory and do the analysis
                            run_analysis = True
                            path_query = "{}/{}/percentile_{}/min_threshold_{}".format(opts.out_dir,m,i_percentile,x)
                            if os.path.exists(path_query):
                                # Is there directory empty?
                                if len(os.listdir(path_query)) > 0: # or os.path.getsize > ?
                                    if opts.force_overwrite == False:
                                        run_analysis = False
                            else:
                                os.makedirs(path_query, exist_ok=True)

                            if run_analysis:
                                # Calculate weights
                                w_ = model.extract_weights(min_threshold=x, name=opts.name, ascending=False)

                                # Are there weights for this threshold?
                                if not w_.isnull().any():
                                    weights_ = w_.to_frame("weights")
                                    weights_.to_csv("{}/weights.tsv.gz".format(path_query), sep="\t", compression="gzip")

                                    # NOTE: This is a hack for outputing the weights for each class but only works for logistic regression.  Once this tree methods are converted to OneVsRest then this can be properly integrated
                                    if m == "logistic":
                                        if len(model.class_labels) > 2:
                                            weights_class_specific_ = model.extract_weights(min_threshold=x, name=opts.name, mode=3)
                                            weights_class_specific_.to_csv("{}/weights.class-specific.tsv.gz".format(path_query), sep="\t", compression="gzip")

                                    log_info.info("\t\t{} | Percentile={} | Minimum threshold={} | Cross validation".format(m,i_percentile,x))
                                    # Synthesize Datasets

                                    if opts.save_data:
                                        path_save = "{}/model-data.pbz2".format(path_query)
                                    else:
                                        path_save = None

                                    scores_ = model.cross_validate(model=clone(estimator),
                                                                    min_threshold=x,
                                                                    cv=opts.cv,
                                                                    cv_labels=cv_labels,
                                                                    method=opts.method,
                                                                    early_stopping=opts.early_stopping,
                                                                    adaptive_range = opts.adaptive_range,
                                                                    adaptive_steps = opts.adaptive_steps,
                                                                    n_jobs=opts.n_jobs,
                                                                    verbose=True,
                                                                    func_groupby=None,
                                                                    path_save=path_save,
                                                                    target_score=best_accuracy_for_modeltype,
                                                                    desc="{} | CV | Percentile={} | Minimum threshold={}".format(m,i_percentile,x),
                                                                    log_file=log_crossvalidation,
                                                                    log_prefix = "{} | pctl={} | t={}".format(m,i_percentile,x)
                                                                    )

                                    log_info.info("\t\t{} | Percentile={} | Minimum threshold={} | Writing scores".format(m,i_percentile,x))
                                    scores_.to_csv("{}/scores.tsv.gz".format(path_query), sep="\t", compression="gzip")

                                    # Update the scores
                                    idx_best_accuracy_for_minthreshold = scores_["accuracy"].sort_values(ascending=False).index[0]
                                    query_accuracy = scores_.loc[idx_best_accuracy_for_minthreshold,"accuracy"]
                                    query_sem = scores_.loc[idx_best_accuracy_for_minthreshold,"sem"]
                                    attribute_set = scores_.loc[idx_best_accuracy_for_minthreshold, "{}_set".format(opts.attr_type)]

                                    # Update accuracies for current percentile
                                    if query_accuracy > best_accuracy_for_percentile:
                                        best_min_threshold_for_percentile = x
                                        if scores_.shape[0] >= 10:
                                            best_results_for_percentile = scores_.sort_values(["accuracy", "sem", "num_{}_included".format(opts.attr_type)], ascending=[False,True, False]).iloc[:10,:3]
                                        else:
                                            best_results_for_percentile = scores_.sort_values(["accuracy", "sem", "num_{}_included".format(opts.attr_type)], ascending=[False,True, False]).iloc[:scores_.shape[0],:3]


                                        best_accuracy_for_percentile = query_accuracy
                                        best_weights_for_percentile = w_.copy()
                                        log_info.info("\t\t{} | Percentile={} | Minimum threshold={} | Updating the model object with accuracy={}".format(m,i_percentile,x,query_accuracy))

                                        # May be slower than necessary to put this here but it's safer
                                        if opts.save_model:
                                            model.save_model(path="{}/{}/percentile_{}/model.pbz2".format(opts.out_dir,m,i_percentile), compression="bz2")

                                        # Update accuracies for current model
                                        if query_accuracy > best_accuracy_for_modeltype:
                                            best_configuration_for_modeltype = (i_percentile, x)
                                            best_hyperparameters_for_modeltype = model.best_hyperparameters_
                                            best_accuracy_for_modeltype = query_accuracy
                                            best_attributes_for_modeltype = attribute_set

                                    log_info.info("\t\t{} | Percentile={} | Minimum threshold={} | Creating plot".format(m,i_percentile,x))

                                    fig, ax = model.plot_scores(title="{} | {} | percentile={} | min_threshold={}".format(opts.name,m,i_percentile,x), ylim=(0,1), min_threshold=x, baseline_score=baseline_score)
                                    fig.savefig("{}/scores.png".format(path_query), dpi=300, format="png", bbox_inches="tight")
                                    plt.close()

                                    # Synthesize summary table
                                    Se_query = pd.Series([m, model.best_hyperparameters_, i_percentile, x, baseline_percentile_0, baseline_score, query_accuracy,  query_sem, query_accuracy - baseline_percentile_0, opts.random_state, len(attribute_set), attribute_set],
                                                   index=["model_type", "hyperparameters", "percentile",  "min_threshold", "baseline_score", "baseline_score_of_current_percentile", "accuracy", "sem", "delta", "random_state", "num_{}_included".format(opts.attr_type), "{}_set".format(opts.attr_type)])
                                    summary_table.append(Se_query)

                                else:

                                    log_info.error("\t\t{} | Percentile={} | Minimum threshold={} | No models had an accuracy for this threshold so the weights were NaN".format(m,i_percentile,x))
                                    if os.path.exists(path_query):
                                        if len(os.listdir(path_query)) == 0:
                                            os.rmdir(path_query)
                                            log_info.error("\t\t{} | Percentile={} | Minimum threshold={} | Removing {}".format(m,i_percentile,x,path_query))
                                    break



                    # Plot weights
                    fig, ax = model.plot_weights(title="{} | {} | percentile={}".format(opts.name,m,i_percentile))
                    fig.savefig("{}/{}/percentile_{}/weights.png".format(opts.out_dir,m,i_percentile), dpi=300, format="png", bbox_inches="tight")
                    plt.close()

                    log_info.info("\t\t{} | Percentile={} | Complete".format(m,i_percentile))

                    print("================\n Results \n================", file=f_summary)
                    print("Best hyperparameters:", model.best_hyperparameters_, sep="\t", file=f_summary)
                    print("Best weights from minimum threshold:", best_min_threshold_for_percentile, 2*"\n", sep="\t", file=f_summary)

                    try:
                        print(best_results.to_string(), file=f_summary)
                    except:
                        if best_results_for_percentile is not None:
                            print("\n".join(str(best_results_for_percentile.iloc[0,:]).split("\n")[:-1]), file=f_summary)
                    print(2*"\n", file=f_summary)


                    # Close summary file
                    f_summary.close()

            # Create next dataset
            if current_iteration+1 < len(opts.percentiles):
                next_percentile = opts.percentiles[current_iteration+1]

                # Get positions
                number_of_attributes_for_next_iteration = int(round((1.0-next_percentile)*n_attributes))

                print(m, current_iteration, i_percentile, next_percentile, opts.percentiles, number_of_attributes_for_next_iteration, opts.min_threshold,  sep="\t")
                # Get attribute set
                assert best_weights_for_percentile is not None, "Try this again with --force_overwrite because certain files are being read incorrectly"
                idx_next_attribute_set = best_weights_for_percentile.nlargest(number_of_attributes_for_next_iteration).index

                # Create filepath
                os.makedirs("{}/{}/percentile_{}".format(opts.out_dir, m,next_percentile), exist_ok=True)
                path_next_attribute_matrix = "{}/{}/percentile_{}/X.subset.pbz2".format(opts.out_dir,m,next_percentile) #Current iteration and not current_itation + 1 because it starts at 1.  It is already offset
                # Check if the file exists already and overwrite if necessary
                update_next_attribute_matrix = True
                if os.path.exists(path_next_attribute_matrix):
                    if opts.force_overwrite == False:
                        update_next_attribute_matrix = False
                if update_next_attribute_matrix:
                    X.loc[:,idx_next_attribute_set].to_pickle(path_next_attribute_matrix, compression="bz2")
                # Update the attributes for next iteration
                path_attributes_current = path_next_attribute_matrix

        # Save best configuration
        print("= == === ===== ======= ============ =====================".replace("=","."), file=sys.stderr)
        print("{} | Percentile={} | Minimum threshold = {} | Baseline score = {} | Best score = {} | ∆ = {} | n_attrs = {}".format(m,best_configuration_for_modeltype[0],best_configuration_for_modeltype[1],to_precision(baseline_percentile_0),to_precision(best_accuracy_for_modeltype),to_precision(best_accuracy_for_modeltype - baseline_percentile_0),len(best_attributes_for_modeltype)), file=sys.stderr)
        print("= == === ===== ======= ============ =====================".replace("=","."), file=sys.stderr)

        # Copy best configuration
        path_src = "{}/{}/percentile_{}".format(opts.out_dir,m,best_configuration_for_modeltype[0])
        path_dst = "{}/{}__pctl_{}__t_{}".format(path_synopsis,m,best_configuration_for_modeltype[0],best_configuration_for_modeltype[1])

        for file_suffix in ["weights.tsv.gz", "scores.tsv.gz", "scores.png", "weights.class-specific.tsv.gz"]:
            src = "{}/min_threshold_{}/{}".format(path_src,best_configuration_for_modeltype[1],file_suffix)
            if os.path.exists(src):
                shutil.copy2(src="{}/min_threshold_{}/{}".format(path_src,best_configuration_for_modeltype[1],file_suffix),
                             dst="{}__{}".format(path_dst,file_suffix),
                )
        # Copy summary file
        for file_suffix in ["acu.tsv.gz", "hyperparameters.tsv.gz", "summary.txt", "weights.png", "weights.tsv.gz"]:
            src = "{}/{}/percentile_{}/{}".format(opts.out_dir,m,best_configuration_for_modeltype[0],file_suffix)
            if os.path.exists(src):
                shutil.copy2(src=src,
                             dst="{}/{}__pctl_{}__{}".format(path_synopsis,m,best_configuration_for_modeltype[0],file_suffix)
                )
        print("", file=sys.stderr)

    # Analysis summary
    df_summarytable = pd.DataFrame(summary_table)
    df_summarytable.index.name = "Iteration"
    df_summarytable.to_csv("{}/{}__synopsis.tsv".format(path_synopsis,opts.name), sep="\t")
    log_info.info("\t\tTotal time:\t {}".format(format_duration(start_time)))
    df_summarytable.to_csv(sys.stdout, sep="\t")

# Initialize
if __name__ == "__main__":
    # Check version
    python_version = sys.version.split(" ")[0]
    condition_1 = int(python_version.split(".")[0]) == 3
    condition_2 = int(python_version.split(".")[1]) >= 6
    assert all([condition_1, condition_2]), "Python version must be >= 3.6.  You are running: {}\n{}".format(python_version,sys.executable)
    main(sys.argv)



# ==============
# Soothsayer
# ==============
# High-level package for (bio-)informatics
# ------------------------------------
# GitHub: https://github.com/jolespin/soothsayer
# PyPI: https://pypi.org/project/soothsayer/
# ------------------------------------
# =======
# Contact
# =======
# Producer: Josh L. Espinoza
# Contact: jespinoz@jcvi.org, jol.espinoz@gmail.com
# Google Scholar: https://scholar.google.com/citations?user=r9y1tTQAAAAJ&hl
# =======
# License BSD-3
# =======
# https://opensource.org/licenses/BSD-3-Clause
#
# Copyright 2018-2020 Josh L. Espinoza
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#
# =======
# Version
# =======
import sys, time
__version__= "2020.12.16"
__moniker__ = "Beskar"
#datetime.datetime.utcnow().strftime("%Y.%m")
__author__ = "Josh L. Espinoza"
__email__ = "jespinoz@jcvi.org, jol.espinoz@gmail.com"
__url__ = "https://github.com/jolespin/soothsayer"
__cite__ = "https://github.com/jolespin/soothsayer"
__license__ = "BSD-3"
__developmental__ = True
t0 = time.time()
# =======
# Shortcuts
# =======
# Core
from .core import core, Dataset
# Classification
from .classification import classification
# Database
from .db import db
# Feature Extraction
from .feature_extraction import feature_extraction
# Hierarchy
from .hierarchy import hierarchy, Agglomerative, Topology
# I/O
from .io import io
# Microbiome
from .microbiome import microbiome
# Networks
from .networks import networks, Hive, TemporalNetwork, EnsembleAssociationNetwork, SampleSpecificPerturbationNetwork, DifferentialEnsembleAssociationNetwork
# Ordination
from .ordination import ordination, PrincipalComponentAnalysis, PrincipalCoordinatesAnalysis, Manifold
# R Wrappers
from .r_wrappers import r_wrappers
# Regression
from .regression import regression
# Statistics
from .statistics import statistics
# Symmetry
from .symmetry import Symmetric
# Transmute
from .transmute import transmute
# Tree
from .tree import tree
# Utilities
from .utils import utils, Chromatic, pv, Suppress
# Visualizations
from .visuals import visuals

# =======
# Direct Exports
# =======
_submodules = ["core", "classification", "db", "feature_extraction", "hierarchy", "io", "microbiome",  "networks", "ordination", "r_wrappers", "regression", "statistics", "symmetry", "transmute", "tree", "utils", "visuals"]
_core = ["Dataset"]
_hierarchy = ["Agglomerative", "Topology"]
_networks = ["Hive", "TemporalNetwork","EnsembleAssociationNetwork", "SampleSpecificPerturbationNetwork", "DifferentialEnsembleAssociationNetwork"]
_ordination = ["PrincipalComponentAnalysis", "PrincipalCoordinatesAnalysis", "Manifold"]
_symmetry = ["Symmetric"]
_utils = ["Chromatic", "Suppress", "pv"]

__all__ = _submodules + _core + _hierarchy + _networks + _ordination + _symmetry + _utils
__all__ = sorted(__all__)

print(utils.format_header("Soothsayer Ecosystem"), file=sys.stderr)
print(" * Soothsayer v{}".format(__version__), file=sys.stderr)
import soothsayer_utils as syu
print(" * Soothsayer Utilities v{}".format(syu.__version__), file=sys.stderr)
import compositional as coda
print(" * Compositional v{}".format(coda.__version__), file=sys.stderr)
import hive_networkx as hx
print(" * Hive NetworkX v{}".format(hx.__version__), file=sys.stderr)
import ensemble_networkx as enx
print(" * Ensemble NetworkX v{}".format(enx.__version__), file=sys.stderr)